/*boundary_blend - interpolate boundaries between successive brain slices and
  then fill in those boundaries by interpolating pixel values.
  Copyright (c) 1993 Matthew Belmonte
  Compile with "cc -O2 -o boundary_blend boundary_blend.c -limage -lm".

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <gl/image.h>
/*image.h includes stdio.h*/
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <malloc.h>
#define PI 3.1416

typedef int Boolean;
#define true 1
#define false 0

short *bottom_slice, *top_slice,		/*brain slices.*/
	*bottom_head, *top_head,		/*corresponding head slices.*/
	**interpolated_slices;			/*the object of the exercise.*/
Boolean *top_slice_pixel_unused;		/*iff pixel not yet touched.*/
int INTFAC,					/*interpolation factor.*/
	WIDTH,					/*length of x axis.*/
	HEIGHT;					/*length of y axis.*/
unsigned short IMAGE_TYPE;			/*keeps IRIS happy.*/
double max_radius_squared;			/*for find_nearest_neighbour.*/

/*Load the given image file into the given memory store.  (orig_file is used
  only for error reporting.)*/
void open_image(store, file, orig_file)
short *store;
char *file, *orig_file;
  {
  register int y;
  IMAGE *image;
  if((image = iopen(file, "r")) == NULL)
    {
    perror((char *)0);
    fprintf(stderr, "can't open image file %s\n", file);
    free(top_head);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  if((image->xsize != WIDTH) || (image->ysize != HEIGHT) || (image->zsize != 1))
    {
    fprintf(stderr,"dimensions of %s (%dx%dx1) and %s (%dx%dx%d) don't match\n",
	orig_file, WIDTH, HEIGHT, file, image->xsize, image->ysize, image->zsize);
    iclose(image);
    free(top_head);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    getrow(image, store + y*WIDTH, y, 0);
  iclose(image);
  }

/*Load the four given image files.  They are the bottom brain slice, the top
  brain slice, the corresponding bottom whole-head slice, and the corresponding
  top whole-head slice, respectively.*/
void load_images(bottom, top, bottomh, toph)
char *bottom, *top, *bottomh, *toph;
  {
  register int y;
  IMAGE *image;
  if((image = iopen(bottom, "r")) == NULL)
    {
    perror((char *)0);
    fprintf(stderr, "can't open image file %s\n", bottom);
    exit(1);
    }
  if(image->zsize != 1)
    {
    fprintf(stderr, "It's not a black-and-white issue.\n");
    iclose(image);
    exit(1);
    }
  IMAGE_TYPE = image->type;
  WIDTH = image->xsize;
  HEIGHT = image->ysize;
  if((bottom_slice = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    iclose(image);
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    getrow(image, bottom_slice + y*WIDTH, y, 0);
  iclose(image);
  if((top_slice = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(bottom_slice);
    exit(1);
    }
  if((bottom_head = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  if((top_head = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  open_image(top_slice, top, bottom);
  open_image(bottom_head, bottomh, bottom);
  open_image(top_head, toph, bottom);
  max_radius_squared = ((double)(WIDTH*HEIGHT))/600.0; /*good for brains*/
  }

void cleanup()
  {
  register int slice;
  for(slice = INTFAC-2; slice >= 0; slice--)
    free(interpolated_slices[slice]);
  free(interpolated_slices);
  free(top_head);
  free(bottom_head);
  free(top_slice);
  free(bottom_slice);
  }

void write_image(image, file)
short *image;
char *file;
  {
  register int y;
  IMAGE *image_file;
  if((image_file = iopen(file, "w", IMAGE_TYPE, 3, WIDTH, HEIGHT, 1)) == NULL)
    {
    cleanup();
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    putrow(image_file, image + y*WIDTH, y, 0);
  iclose(image_file);
  }

Boolean in_bounds(x, y)
int x, y;
  {
  return((x >= 0) && (x < WIDTH) && (y >= 0) && (y < HEIGHT));
  }

short get_pixel(image, x, y)
short *image;
int x, y;
  {
  return(in_bounds(x, y) ? image[y*WIDTH + x] : 0);
  }

Boolean pixel_is_in_image(image, x, y)
short *image;
int x, y;
  {
  return(get_pixel(image, x, y) != 0);
  }

void set_pixel(image, x, y, val)
short *image;
int x, y, val;
  {
  image[y*WIDTH + x] = val;
  }

void clear_pixel(image, x, y)
short *image;
int x, y;
  {
  set_pixel(image, x, y, 0);
  }

void interpolate_pixel(slice, x, y)
int slice, x, y;
  {
  set_pixel(interpolated_slices[slice-1], x, y, (get_pixel(bottom_head, x, y) * slice + INTFAC/2) / INTFAC + (get_pixel(top_head, x, y) * (INTFAC - slice) + INTFAC/2) / INTFAC);
  }

Boolean top_slice_pixel_unmarked(x, y)
int x, y;
  {
  return(top_slice_pixel_unused[WIDTH*y+x]);
  }

void mark_top_slice_pixel(x, y)
int x, y;
  {
  top_slice_pixel_unused[WIDTH*y+x] = false;
  }

void alloc_interpolated_maps()
  {
  register int x, y, slice;
  interpolated_slices = (short **)calloc(INTFAC-1, sizeof(short *));
  if(interpolated_slices == (short **)0)
    {
    perror((char *)0);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  for(slice = 0; slice != INTFAC-1; slice++)
    {
    interpolated_slices[slice] = (short *)calloc(WIDTH * HEIGHT, sizeof(short));
    if(interpolated_slices[slice] == (short *)0)
      {
      perror((char *)0);
      while(--slice >= 0)
	free(interpolated_slices[slice]);
      free(interpolated_slices);
      free(top_slice);
      free(bottom_slice);
      exit(1);
      }
    for(y = 0; y != HEIGHT; y++)
      for(x = 0; x != WIDTH; x++)
	clear_pixel(interpolated_slices[slice], x, y);
    }
  }

/*Find a non-zero pixel on the circumference of the given circle in the
  perimeter map of the top slice.  Return false if all points on this circle
  are zero.*/
Boolean find_point_on_circumference(map, xc, yc, r, x, y)
short *map;
int xc, yc;
double r;
int *x, *y;
  {
  double angle, angle_step;
  angle_step = atan(1.0/r);
  angle = 0.0;
  *x = xc + (int)(r+0.5);
  *y = yc;
  while((angle < 2*PI) && !pixel_is_in_image(map, *x, *y))
    {
    angle += angle_step;
    *x = xc + (int)(r*cos(angle)+0.5);
    *y = yc + (int)(r*sin(angle)+0.5);
    }
  return(angle < 2*PI);
  }

/*Set (neighbouring_x, neighbouring_y) to the point in the given map whose
  location most nearly corresponds to (x, y), or return false if there is no
  such point in the neighbourhood.*/
Boolean find_nearest_neighbour(map, x, y, neighbouring_x, neighbouring_y)
short *map;
int x, y, *neighbouring_x, *neighbouring_y;
  {
  static short *cached_map;
  static int cached_x = -1, cached_y = -1;
  static double radius_squared;
  if(pixel_is_in_image(map, x, y))
    {
    *neighbouring_x = x;
    *neighbouring_y = y;
    radius_squared = 0.0;
    }
  else
    {
    if((map == cached_map) && (abs(x - cached_x) + abs(y - cached_y) <= 1))
      radius_squared = radius_squared - 2.0*sqrt(radius_squared) + 1.0;
    else
      radius_squared = 1.0;
    /*inv: all pixels inside the circle are black*/
    while((radius_squared < max_radius_squared) && !find_point_on_circumference(map, x, y, sqrt(radius_squared), neighbouring_x, neighbouring_y))
      radius_squared += 1.0;
    }
  cached_map = map;
  cached_x = x;
  cached_y = y;
  return(radius_squared < max_radius_squared);
  }

/*The only differences between the paint and the mark functions are that the
  paint function assumes that the area to be painted is closed, and the mark
  function fills the area with -1's instead of interpolated grey levels.  The
  mark function is applied to mark the exterior of the image, so that the paint
  function can be applied to the remaining black areas, which have remained
  unmarked and therefore are closed.*/

void mark_vertical(int, int, int, int);

void mark_horizontal(slice, x, y, dx)
int slice, x, y, dx;
  {
  while(in_bounds(x, y) && !pixel_is_in_image(interpolated_slices[slice-1], x, y))
    {
    set_pixel(interpolated_slices[slice-1], x, y, -1);
    mark_vertical(slice, x, y+1, 1);
    mark_vertical(slice, x, y-1, -1);
    x += dx;
    }
  }

void mark_vertical(slice, x, y, dy)
int slice, x, y, dy;
  {
  while(in_bounds(x, y) && !pixel_is_in_image(interpolated_slices[slice-1], x, y))
    {
    set_pixel(interpolated_slices[slice-1], x, y, -1);
    mark_horizontal(slice, x-1, y, -1);
    mark_horizontal(slice, x+1, y, 1);
    y += dy;
    }
  }

void paint_vertical(int, int, int, int);

void paint_horizontal(slice, x, y, dx)
int slice, x, y, dx;
  {
  while(!pixel_is_in_image(interpolated_slices[slice-1], x, y))
    {
    interpolate_pixel(slice, x, y);
    paint_vertical(slice, x, y+1, 1);
    paint_vertical(slice, x, y-1, -1);
    x += dx;
    }
  }

void paint_vertical(slice, x, y, dy)
int slice, x, y, dy;
  {
  while(!pixel_is_in_image(interpolated_slices[slice-1], x, y))
    {
    interpolate_pixel(slice, x, y);
    paint_horizontal(slice, x-1, y, -1);
    paint_horizontal(slice, x+1, y, 1);
    y += dy;
    }
  }

void paint(slice, x, y)
int slice, x, y;
  {
  paint_horizontal(slice, x, y, -1);
  paint_horizontal(slice, x+1, y, 1);
  }

void main(argc, argv)
int argc;
char *argv[];
  {
  register int slice;
  int bx, tx, dx, by, ty, dy;			/*bottom, top, difference*/
  if(argc < 6)
    {
    fprintf(stderr, "usage: %s brain1 brain2 head1 head2 outimage1 ... outimageN\n", argv[0]);
    exit(1);
    }
  INTFAC = argc - 4;
  load_images(argv[1], argv[2], argv[3], argv[4]);
  alloc_interpolated_maps();
  if((top_slice_pixel_unused = (Boolean *)calloc(WIDTH*HEIGHT, sizeof(Boolean))) == (Boolean *)0)
    {
    cleanup();
    exit(1);
    }
  for(ty = 0; ty != HEIGHT; ty++)
    for(tx = 0; tx != WIDTH; tx++)
      top_slice_pixel_unused[ty*WIDTH+tx] = pixel_is_in_image(top_slice, tx, ty);
  for(by = 0; by != HEIGHT; by++)
    for(bx = 0; bx != WIDTH; bx++)
      if(pixel_is_in_image(bottom_slice, bx, by) && find_nearest_neighbour(top_slice, bx, by, &tx, &ty))
	{
	dx = tx - bx;
	dy = ty - by;
	for(slice = 1; slice != INTFAC; slice++)
	  interpolate_pixel(slice, bx + (dx*slice+INTFAC/2)/INTFAC, by + (dy*slice+INTFAC/2)/INTFAC);
	mark_top_slice_pixel(tx, ty);
	}
  for(ty = 0; ty != HEIGHT; ty++)
    for(tx = 0; tx != WIDTH; tx++)
      if(top_slice_pixel_unmarked(tx, ty) && find_nearest_neighbour(bottom_slice, tx, ty, &bx, &by))
	{
	dx = tx - bx;
	dy = ty - by;
	for(slice = 1; slice != INTFAC; slice++)
	  interpolate_pixel(slice, bx + (dx*slice+INTFAC/2)/INTFAC, by + (dy*slice+INTFAC/2)/INTFAC);
	}
  free(top_slice_pixel_unused);
  for(slice = 1; slice != INTFAC; slice++)
    {
    mark_horizontal(slice, 0, 0, 1);
    for(ty = 0; ty != HEIGHT; ty++)
      for(tx = 0; tx != WIDTH; tx++)
        if(!pixel_is_in_image(interpolated_slices[slice-1], tx, ty))
          paint(slice, tx, ty);
    for(ty = 0; ty != HEIGHT; ty++)
      for(tx = 0; tx != WIDTH; tx++)
	if(get_pixel(interpolated_slices[slice-1], tx, ty) == -1)
	  clear_pixel(interpolated_slices[slice-1], tx, ty);
    write_image(interpolated_slices[slice-1], argv[slice+4]);
    }
  cleanup();
  exit(0);
  }
