/*boundary_blend - interpolate boundaries between successive brain slices and
  then fill in those boundaries by interpolating pixel values.

  Copyright (c) 1993 Matthew Belmonte

  This is somewhat of a hack; make sure that there's at least one pixel's width
  of black boundary around the edge of the image, or the world will fall apart.
  Compile with "cc -O2 -o boundary_blend boundary_blend.c -limage".

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include "/usr/people/4Dgifts/iristools/include/image.h"
/*image.h includes stdio.h*/
#include <math.h>
#include <sys/types.h>
#include <malloc.h>
#define PI 3.14159

typedef int Boolean;
#define true 1
#define false 0
typedef struct {int first, second;} PAIR;

short *bottom_slice, *top_slice,		/*brain slices.*/
	*bottom_head, *top_head,		/*corresponding head slices.*/
	*bottom_perimeter, *top_perimeter,	/*generated from brain slices.*/
	**interpolated_slices;			/*the object of the exercise.*/
int INTFAC,					/*interpolation factor.*/
	WIDTH,					/*length of x axis.*/
	HEIGHT;					/*length of y axis.*/
unsigned short IMAGE_TYPE;			/*keeps IRIS happy.*/

/*Load the given image file into the given memory store.  (orig_file is used
  only for error reporting.)*/
void open_image(store, file, orig_file)
short *store;
char *file, *orig_file;
  {
  register int y;
  IMAGE *image;
  if((image = iopen(file, "r")) == NULL)
    {
    perror((char *)0);
    fprintf(stderr, "can't open image file %s\n", file);
    free(top_head);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  if((image->xsize != WIDTH) || (image->ysize != HEIGHT) || (image->zsize != 1))
    {
    fprintf(stderr,"dimensions of %s (%dx%dx1) and %s (%dx%dx%d) don't match\n",
	orig_file, WIDTH, HEIGHT, file, image->xsize, image->ysize, image->zsize);
    iclose(image);
    free(top_head);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    getrow(image, store + y*WIDTH, y, 0);
  iclose(image);
  }

/*Load the four given image files.  They are the bottom brain slice, the top
  brain slice, the corresponding bottom whole-head slice, and the corresponding
  top whole-head slice, respectively.*/
void load_images(bottom, top, bottomh, toph)
char *bottom, *top, *bottomh, *toph;
  {
  register int y;
  IMAGE *image;
  if((image = iopen(bottom, "r")) == NULL)
    {
    perror((char *)0);
    fprintf(stderr, "can't open image file %s\n", bottom);
    exit(1);
    }
  if(image->zsize != 1)
    {
    fprintf(stderr, "It's not a black-and-white issue.\n");
    iclose(image);
    exit(1);
    }
  IMAGE_TYPE = image->type;
  WIDTH = image->xsize;
  HEIGHT = image->ysize;
  if((bottom_slice = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    iclose(image);
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    getrow(image, bottom_slice + y*WIDTH, y, 0);
  iclose(image);
  if((top_slice = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(bottom_slice);
    exit(1);
    }
  if((bottom_head = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  if((top_head = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(bottom_head);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  open_image(top_slice, top, bottom);
  open_image(bottom_head, bottomh, bottom);
  open_image(top_head, toph, bottom);
  }

void cleanup()
  {
  register int slice;
  for(slice = INTFAC-2; slice >= 0; slice--)
    free(interpolated_slices[slice]);
  free(interpolated_slices);
  free(top_perimeter);
  free(bottom_perimeter);
  free(top_head);
  free(bottom_head);
  free(top_slice);
  free(bottom_slice);
  }

void write_image(image, file)
short *image;
char *file;
  {
  register int y;
  IMAGE *image_file;
  if((image_file = iopen(file, "w", IMAGE_TYPE, 3, WIDTH, HEIGHT, 1)) == NULL)
    {
    cleanup();
    exit(1);
    }
  for(y = 0; y != HEIGHT; y++)
    putrow(image_file, image + y*WIDTH, y, 0);
  iclose(image_file);
  }

short get_pixel(image, x, y)
short *image;
int x, y;
  {
  return(image[y*WIDTH + x]);
  }

void set_pixel(image, x, y, val)
short *image;
int x, y, val;
  {
  image[y*WIDTH + x] = val;
  }

Boolean pixel_is_in_image(image, x, y)
short *image;
int x, y;
  {
  return(get_pixel(image, x, y) != 0);
  }

void interpolate_pixel(slice, x, y)
int slice, x, y;
  {
  set_pixel(interpolated_slices[slice-1], x, y, (get_pixel(bottom_head, x, y) * slice + INTFAC/2) / INTFAC + (get_pixel(top_head, x, y) * (INTFAC - slice) + INTFAC/2) / INTFAC);
  }

void alloc_interpolated_maps()
  {
  register int x, y, slice;
  interpolated_slices = (short **)calloc(INTFAC-1, sizeof(short *));
  if(interpolated_slices == (short **)0)
    {
    perror((char *)0);
    free(top_perimeter);
    free(bottom_perimeter);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  for(slice = 0; slice != INTFAC-1; slice++)
    {
    interpolated_slices[slice] = (short *)calloc(WIDTH * HEIGHT, sizeof(short));
    if(interpolated_slices[slice] == (short *)0)
      {
      perror((char *)0);
      while(--slice >= 0)
	free(interpolated_slices[slice]);
      free(interpolated_slices);
      free(top_perimeter);
      free(bottom_perimeter);
      free(top_slice);
      free(bottom_slice);
      exit(1);
      }
    for(y = 0; y != HEIGHT; y++)
      for(x = 0; x != WIDTH; x++)
	set_pixel(interpolated_slices[slice], x, y, 0);
    }
  }

/*Generate a perimeter map of src_image.  Store this map in obj_image.*/
void find_perimeters(src_image, obj_image)
short *src_image, *obj_image;
  {
  register int x, y, pixel0, pixel1;
  for(y = 0; y != WIDTH; y++)
    set_pixel(obj_image, 0, y, 0);
  for(y = 0; y != HEIGHT; y++)
    for(x = 1; x != WIDTH; x++)
      {
      pixel0 = get_pixel(src_image, x-1, y);
      pixel1 = get_pixel(src_image, x, y);
      if((pixel0 == 0) && (pixel1 != 0))
	set_pixel(obj_image, x, y, pixel1);
      else
	{
	if((pixel0 != 0) && (pixel1 == 0))
	  set_pixel(obj_image, x-1, y, pixel0);
	set_pixel(obj_image, x, y, 0);
	}
      }
  for(x = 0; x != WIDTH; x++)
    for(y = 1; y != HEIGHT; y++)
      {
      pixel0 = get_pixel(src_image, x, y-1);
      pixel1 = get_pixel(src_image, x, y);
      if((pixel0 == 0) && (pixel1 != 0))
	set_pixel(obj_image, x, y, pixel1);
      else if((pixel0 != 0) && (pixel1 == 0))
	set_pixel(obj_image, x, y-1, pixel0);
      }
  }

/*Choose a pixel in the perimeter map of the bottom slice that has not yet been
  processed.*/
Boolean choose_perimeter_point(point)
PAIR *point;
  {
  static int x=0, y=0;
  while((y != HEIGHT) && (!pixel_is_in_image(bottom_perimeter, x, y)))
    if(++x == WIDTH)
      {
      x = 0;
      y++;
      }
  point->first = x;
  point->second = y;
  return(y != HEIGHT);
  }

/*Find a non-zero pixel on the circumference of the given circle in the
  perimeter map of the top slice.  Return false if all points on this circle
  are zero.*/
Boolean find_point_on_circumference(xc, yc, r, result)
int xc, yc, r;
PAIR *result;
  {
  register int x, y;
  double angle, angle_step;
  angle_step = PI / (4*r);
  angle = 0.0;
  x = xc + r;
  y = yc;
  while((angle < 2*PI) && !pixel_is_in_image(top_perimeter, x, y))
    {
    angle += angle_step;
    x = xc + (int)(r*cos(angle)+0.5);
    y = yc + (int)(r*sin(angle)+0.5);
    }
  if(angle >= 2*PI)
    return(false);
  else
    {
    result->first = x;
    result->second = y;
    return(true);
    }
  }

/*bottom_point is some point in the perimeter map of the bottom slice.  Set
  top_point to the point in the perimeter map of the top slice whose location
  most nearly corresponds to that of bottom_point.*/
void find_nearest_neighbour(bottom_point, top_point)
PAIR *bottom_point, *top_point;
  {
  register int radius;
  if(pixel_is_in_image(top_perimeter, bottom_point->first, bottom_point->second))
    *top_point = *bottom_point;
  else
    {
    radius = 1;
    while(!find_point_on_circumference(bottom_point->first, bottom_point->second, radius, top_point))
      radius++;
    }
  }

int max(a,b)
int a,b;
  {
  return(a > b ? a : b);
  }

int abs(a)
int a;
  {
  return(a < 0 ? -a : a);
  }

int hv_dist(x0, y0, x1, y1)
int x0, y0, x1, y1;
  {
  return(max(abs(x1-x0),abs(y1-y0)));
  }

/*Find the next point on the given perimeter, or return false if the perimeter
  ends.  When this function is first called, the current point can be any point
  on the perimeter, and the previous point should be set identical to the
  current point.  On reaching the end of a perimeter, this function returns
  false.*/
Boolean next_perimeter_point(map, prev_x, prev_y, cur_x, cur_y, result)
short *map;
int prev_x, prev_y, cur_x, cur_y;
PAIR *result;
  {
  static int d[8][2] = {{-1,-1},{-1,1},{1,-1},{1,1},{-1,0},{1,0},{0,-1},{0,1}};
  register int i, cur_distance, max_distance;
  result->first = -1;
  max_distance = 0;
  for(i = 0; i != 8; i++)
    if(!((cur_x+d[i][0] == prev_x) && (cur_y+d[i][1] == prev_y)) && pixel_is_in_image(map, cur_x+d[i][0], cur_y+d[i][1]) && ((cur_distance = hv_dist(cur_x+d[i][0], cur_y+d[i][1], prev_x, prev_y)) >= max_distance))
      {
      max_distance = cur_distance;
      result->first = cur_x+d[i][0];
      result->second = cur_y+d[i][1];
      }
  return(result->first != -1);
  }

double distance(x0, y0, z0, x1, y1, z1)
int x0, y0, z0, x1, y1, z1;
  {
  return(sqrt((double)((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0) + (z1-z0)*(z1-z0))));
  }

/*Stretch a rubber band from the given hinge on the bottom perimeter to a point
  on the top perimeter that locally (within the next 200 perimeter points,
  anyway) minimises the stretching of the rubber
  band.  Return this minimising top point in the "result" parameter.*/
void find_local_distance_minimum(hinge, prev_top, cur_top, result)
PAIR *hinge, prev_top, cur_top, *result;
  {
  register int s;
  double min_distance, cur_distance;
  PAIR next_top;
  min_distance = HUGE_VAL;
  for(s = 0; s != 200; s++)
    {
    cur_distance = distance(hinge->first, hinge->second, 0, cur_top.first, cur_top.second, INTFAC);
    if(cur_distance < min_distance)
      {
      min_distance = cur_distance;
      *result = cur_top;
      }
    if(!next_perimeter_point(top_perimeter, prev_top.first, prev_top.second, cur_top.first, cur_top.second, &next_top))
      break;
    prev_top = cur_top;
    cur_top = next_top;
    }
  }

/*Swing a straight-edge hinged at the given location on the bottom perimeter
  from start_point to end_point on the top perimeter.  prev_point neighbours
  start_point and defines the direction of motion along the curve.  Points are
  deleted from the top perimeter as they are passed, and the perimeters of the
  intervening slices are interpolated.*/
void swing_ruler(hinge, prev_point, start_point, end_point)
PAIR *hinge, prev_point, *start_point, *end_point;
  {
  PAIR cur_point, next_point;
  register int slice, dx, dy;
  cur_point = *start_point;
  if((start_point->first == end_point->first) && (start_point->second == end_point->second))
    prev_point.first = -1; /*always do at least one loop, to ensure no gaps*/
  while(!((prev_point.first == end_point->first) && (prev_point.second == end_point->second)))
    {
    dx = cur_point.first - hinge->first;
    dy = cur_point.second - hinge->second;
    for(slice = 1; slice != INTFAC; slice++)
      interpolate_pixel(slice, hinge->first + (dx*slice+INTFAC/2)/INTFAC, hinge->second + (dy*slice+INTFAC/2)/INTFAC);
    set_pixel(top_perimeter, cur_point.first, cur_point.second, 0);
    next_perimeter_point(top_perimeter, prev_point.first, prev_point.second, cur_point.first, cur_point.second, &next_point);
    prev_point = cur_point;
    cur_point = next_point;
    }
  set_pixel(bottom_perimeter, hinge->first, hinge->second, 0);
  }

/*Set the boundary condition so that next_perimeter_point processes in the same
  direction around the top and bottom curves.*/
void set_prev_top_point(b, t, result)
PAIR *b, *t, *result;
  {
  register int b_count, t_count;
  PAIR prev_b, cur_b, next_b, prev_t, cur_t, next_t;
  cur_b = *b;
  cur_t = *t;
  prev_b = cur_b;
  for(b_count = 0; (b_count != 10) && next_perimeter_point(bottom_perimeter, prev_b.first, prev_b.second, cur_b.first, cur_b.second, &next_b); b_count++)
    {
    prev_b = cur_b;
    cur_b = next_b;
    }
  prev_t = cur_t;
  for(t_count = 0; (t_count != 10) && next_perimeter_point(top_perimeter, prev_t.first, prev_t.second, cur_t.first, cur_t.second, &next_t); t_count++)
    {
    prev_t = cur_t;
    cur_t = next_t;
    }
  if(distance(cur_b.first, cur_b.second, 0, cur_t.first, cur_t.second, 0) <= (b_count + t_count)/2.0)
    *result = *t;
  else
    next_perimeter_point(top_perimeter, t->first, t->second, t->first, t->second, result);
  }

void interpolate_boundaries()
  {
  register int loop_counter_kludge;
  PAIR prev_bottom_point, prev_top_point, cur_bottom_point, cur_top_point, next_bottom_point, next_top_point;
  loop_counter_kludge = 0;
  while((loop_counter_kludge++ != 100) && choose_perimeter_point(&cur_bottom_point))
    {
    find_nearest_neighbour(&cur_bottom_point, &cur_top_point);
    prev_bottom_point = cur_bottom_point;
    set_prev_top_point(&cur_bottom_point, &cur_top_point, &prev_top_point);
    while(next_perimeter_point(bottom_perimeter, prev_bottom_point.first, prev_bottom_point.second, cur_bottom_point.first, cur_bottom_point.second, &next_bottom_point))
      {
      find_local_distance_minimum(&next_bottom_point, prev_top_point, cur_top_point, &next_top_point);
      swing_ruler(&next_bottom_point, prev_top_point, &cur_top_point, &next_top_point);
      prev_bottom_point = cur_bottom_point;
      cur_bottom_point = next_bottom_point;
      prev_top_point = cur_top_point;
      cur_top_point = next_top_point;
      }
    }
  }

Boolean pixel_is_corner(slice, x, y)
int slice, x, y;
  {
  return((pixel_is_in_image(interpolated_slices[slice-1], x, y+1) || pixel_is_in_image(interpolated_slices[slice-1], x, y-1))
     && (pixel_is_in_image(interpolated_slices[slice-1], x+1, y) || pixel_is_in_image(interpolated_slices[slice-1], x-1, y)));
  }

void paint_vertical(int, int, int, int);

void paint_horizontal(slice, x, y, dx)
int slice, x, y, dx;
  {
  int x0;
  x0 = x;
  while((x > 0) && (x < WIDTH) && (y > 0) && (y < HEIGHT) && !pixel_is_in_image(interpolated_slices[slice-1], x, y))
    x += dx;
  if((x > 0) && (x < WIDTH) && (y > 0) && (y < HEIGHT))
    {
    x = x0;
    while(!pixel_is_in_image(interpolated_slices[slice-1], x, y))
      {
      interpolate_pixel(slice, x, y);
      paint_vertical(slice, x, y+1, 1);
      paint_vertical(slice, x, y-1, -1);
      x += dx;
      }
    }
  }

void paint_vertical(slice, x, y, dy)
int slice, x, y, dy;
  {
  int y0;
  y0 = y;
  while((x > 0) && (x < WIDTH) && (y > 0) && (y < HEIGHT) && !pixel_is_in_image(interpolated_slices[slice-1], x, y))
    y += dy;
  if((x > 0) && (x < WIDTH) && (y > 0) && (y < HEIGHT))
    {
    y = y0;
    while(!pixel_is_in_image(interpolated_slices[slice-1], x, y))
      {
      interpolate_pixel(slice, x, y);
      paint_horizontal(slice, x-1, y, -1);
      paint_horizontal(slice, x+1, y, 1);
      y += dy;
      }
    }
  }

void paint(slice, x, y)
int slice, x, y;
  {
  paint_horizontal(slice, x, y, -1);
  paint_horizontal(slice, x+1, y, 1);
  }

void interpolate_grey_levels()
  {
  register int x, y, slice;
  for(y = 0; y != HEIGHT; y++)
    for(x = 0; x != WIDTH; x++)
      if(pixel_is_in_image(bottom_slice, x, y) && pixel_is_in_image(top_slice, x, y))
	for(slice = 1; slice != INTFAC; slice++)
	  if(pixel_is_corner(slice, x, y))
	    interpolate_pixel(slice, x, y);
  for(y = 0; y != HEIGHT; y++)
    for(x = 0; x != WIDTH; x++)
      if(pixel_is_in_image(bottom_slice, x, y) && pixel_is_in_image(top_slice, x, y))
	for(slice = 1; slice != INTFAC; slice++)
	  if(!pixel_is_in_image(interpolated_slices[slice-1], x, y))
	    paint(slice, x, y);
  }

void main(argc, argv)
int argc;
char *argv[];
  {
  int slice;
  if(argc < 6)
    {
    fprintf(stderr, "usage: %s brain1 brain2 head1 head2 outimage1 ... outimageN\n", argv[0]);
    exit(1);
    }
  INTFAC = argc - 4;
  load_images(argv[1], argv[2], argv[3], argv[4]);
  if((bottom_perimeter = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  if((top_perimeter = (short *)calloc(WIDTH * HEIGHT, sizeof(short))) == (short *)0)
    {
    perror((char *)0);
    free(bottom_perimeter);
    free(top_slice);
    free(bottom_slice);
    exit(1);
    }
  find_perimeters(bottom_slice, bottom_perimeter);
  find_perimeters(top_slice, top_perimeter);
  alloc_interpolated_maps();
  interpolate_boundaries();
  interpolate_grey_levels();
  for(slice = 0; slice != INTFAC-1; slice++)
    write_image(interpolated_slices[slice], argv[slice+5]);
  cleanup();
  exit(0);
  }
