/*unwarp.c - unwarp a distorted echo-planar magnetic resonance image.
  Copyright (c) 1996 Matthew Belmonte

  Originally written for use with Greg Allen's cerebellar fMRI data:
	Greg Allen, Rick Buxton, Eric Wong, Eric Courchesne,
	`Attentional Activation of the Cerebellum Independent of Motor
	Involvement', Science 275:5308:1940-1943 (1997).

  Compile with "cc -O2 -o unwarp unwarp.c -limage".

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <gl/image.h>

#define EPI_MARGIN 8	/*# of pixels to crop from each edge to equalise FOV*/
#define EPI_WIDTH 64	/*width in pixels of EPI image*/
#define EPI_HEIGHT 64	/*height in pixels of EPI image*/
#define SPGR_WIDTH 256	/*width of SPGR, zoomed EPI, and correlation images*/
#define SPGR_HEIGHT 256	/*height of SPGR, zoomed EPI, and correlation images*/

/*names of output images*/
static char epi_file[] = "epi.bw",
	    zoomed_epi_file[] = "zoomed_epi.bw",
	    unwarped_epi_file[] = "unwarped_epi.bw",
	    unwarped_overlay_file[] = "unwarped_overlay.bw";

static char izoom[] = "/usr/NeWS/bin/izoom";

/*Assuming that pixel values in the intervals [minpix..maxpix] are
  predominantly brain, calculate the centroid of the given row.  That is,
  compute the horizontal offset of the vertical axis of symmetry.  If enough
  data points are found to compute this offset reliably, then subtract the
  'meridian' offset and concatenate the resulting value to the vector s[],
  concatenate an error measure to sd[], and increment the vector length *n.*/
void shift_row(image, minpix, maxpix, meridian, minskew, maxskew, s, sd, n)
unsigned short *image, minpix, maxpix;
float meridian, minskew, maxskew, *s, *sd;
int *n;
  {
  register int x, xsum, xcount;
  xsum = 0;
  xcount = 0;
  for(x = 0; x != SPGR_WIDTH; x++)
    {
    if((image[x] >= minpix) && (image[x] <= maxpix))
      {
      xsum += x;
      xcount++;
      }
    if((image[x] >= minpix) && (image[x] <= maxpix))
      {
      xsum += x;
      xcount++;
      }
    }
  if(xcount > 1)
    {
    *s = xsum/(float)xcount-meridian;
    if((*s >= minskew) && (*s <= maxskew))
      {
      *sd = 1.0/(float)(xcount*xcount);
      (*n)++;
      }
    }
  }

static void swap(v, i, j)
float *v;
int i, j;
  {
  register float temp;
  temp = v[i];
  v[i] = v[j];
  v[j] = temp;
  }

/*Find the median of v[0..len-1].  This is accomplished using a partial
  quicksort; thus the contents of v end up as a permutation of the original
  sequence.*/
float median(v, len)
float *v;
int len;
  {
  register int h, k, m, n;
  m = 0;
  n = len-1;
  do
  /*inv: m <= len/2 < n and v[m..n] is a permutation of the subsequence v[m..n]
    of sorted sequence v[0..len-1].*/
    {
    h = m+1;
    k = n;
    /*inv: v[m+1..h-1] <= v[m] and v[k+1..n] > v[m]*/
    while(h <= k)
      {
      while((h <= k) && (v[h] <= v[m]))
	h++;
      while((h <= k) && (v[k] > v[m]))
	k--;
      if(h < k)
	swap(v,  h++, k--);
      }
    /*k=h-1*/
    if(k==m)
      m++;
    else if(h==n+1)
      swap(v, m, n--);
    else
      {
      swap(v, m, k);
      if(k <= len/2)
	m=k;
      else
	n=k;
      }
    /*v[k] is now in sorted position*/
    } while(k != len/2);
  return(v[k]);
  }

/*3-point smooth v[0..n-1] with boundary conditions v[-1]=0 and v[n]=v[n-1].
  Return 0 if the resulting vector is nondecreasing, 1 otherwise.*/
int smooth(v, n)
float *v;
int n;
  {
  register int i, bump_flag;
  register float sum, temp, temp2;
  bump_flag = 0;
  sum = v[0]+v[1];
  temp = (v[0]+v[1])/3.0;
  for(i = 1; i != n-1; i++)
  /*inv: v[0..i-2] has been 3-point smoothed with boundary condition v[-1]=0,
    sum=v[i-1]+v[i], and temp is the smoothed value of v[i-1].*/
    {
    sum += v[i+1];
    temp2 = sum/3.0;
    bump_flag = (bump_flag || (temp2 < temp));
    sum -= v[i-1];
    v[i-1] = temp;
    temp = temp2;
    }
  /*post: v[0..n-3] has been 3-point smoothed with boundary condition v[-1]=0,
    sum=v[n-2]+v[n-1], and temp is the smoothed value of v[n-2].*/
  v[n-2] = temp;
  v[n-1] = (sum+v[n-1])/3.0;
  return(bump_flag || (v[n-1]<v[n-2]));
  }

/*rotate the given row 'skew' pixels to the right*/
void correct_row(row, skew)
unsigned short *row;
int skew;
  {
  register int x;
  unsigned short temprow[SPGR_WIDTH];
  if(skew > 0)
    {
    for(x = 0; x != SPGR_WIDTH; x++)
      temprow[(x+skew)%SPGR_WIDTH] = row[x];
    for(x = 0; x != SPGR_WIDTH; x++)
      row[x] = temprow[x];
    }
  }

/*correct all rows in the image, by interpolating between successive
  skew values*/
void correct_image(image, slice_num, skew, n)
unsigned short image[][SPGR_WIDTH];
float *slice_num, *skew;
int n;
  {
  register int i, y, prev_slice;
  register float prev_skew;
  y = -1;
  prev_slice = 0;
  prev_skew = 0.0;
  for(i = 0; i != n; i++)
    while(y != (int)(slice_num[i]))
      {
      y++;
      correct_row(image[y], (int)(0.5+prev_skew+(skew[i]-prev_skew)*(y-prev_slice)/(slice_num[i]-prev_slice)));
      }
  while(y != SPGR_HEIGHT)
    correct_row(image[y++], (int)(0.5+skew[n-1]));
  }

/*read the named image into the given buffer*/
void read_spgr_image(filename, imgbuf)
char *filename;
unsigned short imgbuf[][SPGR_WIDTH];
  {
  register int y;
  IMAGE *img;
  img = iopen(filename, "r");
  if(img == (IMAGE *)0)
    {
    fprintf(stderr, "Couldn't open %s\n", filename);
    exit(1);
    }
  if((img->xsize != SPGR_WIDTH) || (img->ysize != SPGR_HEIGHT))
    {
    iclose(img);
    fprintf(stderr, "%dx%d image %s is of the wrong size\n", img->xsize, img->ysize, filename);
    exit(1);
    }
  for(y = 0; y != SPGR_HEIGHT; y++)
    getrow(img, imgbuf[y], y, 0);
  iclose(img);
  }

/*write the named image from the given buffer*/
void write_spgr_image(filename, imgbuf)
char *filename;
unsigned short imgbuf[][SPGR_WIDTH];
  {
  register int y;
  IMAGE *img;
  img = iopen(filename, "w", RLE(1), 3, SPGR_WIDTH, SPGR_HEIGHT, 1);
  for(y = 0; y != SPGR_HEIGHT; y++)
    putrow(img, imgbuf[y], y, 0);
  iclose(img);
  }

void main(argc, argv)
int argc;
char **argv;
  {
  unsigned short epi_value;
  unsigned long max_epi_value;
  int shift, trial, slice, image, x, y, minpix, maxpix;
  FILE *fp;
  IMAGE *img;
  char horizontal_zoom[12], vertical_zoom[12], filename[MAXPATHLEN];
  unsigned long epi[EPI_HEIGHT-2*EPI_MARGIN][EPI_WIDTH-2*EPI_MARGIN];
  unsigned short rowbuf[EPI_WIDTH-2*EPI_MARGIN], spgr[SPGR_HEIGHT][SPGR_WIDTH], zoomed_epi[SPGR_HEIGHT][SPGR_WIDTH];
  float vertical_meridian, slice_num[SPGR_HEIGHT], skew[SPGR_HEIGHT], stdev[SPGR_HEIGHT];
  int num_points;
  printf("unwarp (C) 1996 by Matthew Belmonte - use requires proper acknowledgement.\n");
/*check and convert arguments*/
  if((argc != 6) && (argc != 8))
    {
    fprintf(stderr, "usage: %s <trial> <slice> <EPI directory> <SPGR image> <warped overlay image> [min. brain pixel] [max. brain pixel]\n", *argv);
    exit(1);
    }
  maxpix = ((argc == 8)? atoi(argv[7]): 89);
  minpix = ((argc == 8)? atoi(argv[6]): 64);
  trial = atoi(argv[1]);
  slice = atoi(argv[2]);
/*read the SPGR image*/
  read_spgr_image(argv[4], spgr);
/*clear the averaged image*/
  for(y = 0; y != EPI_HEIGHT-2*EPI_MARGIN; y++)
    for(x = 0; x != EPI_WIDTH-2*EPI_MARGIN; x++)
      epi[y][x] = 0L;
/*average all available EPI images*/
  image = 0;
  while((fp = (sprintf(filename, "%s/trial%d.%02d.%04d", argv[3], trial, slice, ++image), fopen(filename, "r"))) != NULL)
    {
  /*skip left margin*/
    fseek(fp, (long)(EPI_WIDTH*EPI_MARGIN*sizeof(short)), SEEK_SET);
    for(x = 0; x != EPI_WIDTH-2*EPI_MARGIN; x++)
      {
    /*skip bottom margin*/
      fseek(fp, (long)(EPI_MARGIN*sizeof(short)), SEEK_CUR);
      for(y = 0; y != EPI_HEIGHT-2*EPI_MARGIN; y++)
	{
	if(fread(&epi_value, sizeof(short), 1, fp) != 1)
	  {
	  fclose(fp);
	  fprintf(stderr, "Error: %s is incomplete\n", filename);
	  exit(1);
	  }
	epi[y][x] += epi_value;
	}
    /*skip top margin*/
      fseek(fp, (long)(EPI_MARGIN*sizeof(short)), SEEK_CUR);
      }
  /*skip right margin*/
    fclose(fp);
    }
  printf("Read %d images from trial %d, slice %d in directory %s\n", --image, trial, slice, argv[3]);
  if(image != 0)
    {
  /*output averaged image*/
    max_epi_value = 0L;
    for(y = 0; y != EPI_HEIGHT-2*EPI_MARGIN; y++)
      for(x = 0; x != EPI_WIDTH-2*EPI_MARGIN; x++)
	if(epi[y][x] > max_epi_value)
	  max_epi_value = epi[y][x];
    for(shift = -8; max_epi_value != 0L; shift++)
      max_epi_value >>= 1;
    if(shift < 0)
      shift = 0;
    img = iopen(epi_file, "w", RLE(1), 3, EPI_WIDTH-2*EPI_MARGIN, EPI_HEIGHT-2*EPI_MARGIN, 1);
    for(y = 0; y != EPI_HEIGHT-2*EPI_MARGIN; y++)
      {
      for(x = 0; x != EPI_WIDTH-2*EPI_MARGIN; x++)
	rowbuf[x] = (unsigned short)(epi[y][x]>>shift);
      putrow(img, rowbuf, y, 0);
      }
    iclose(img);
  /*zoom the cropped EPI image to match the resolution of the SPGR image*/
    sprintf(horizontal_zoom, "%f", SPGR_WIDTH/(float)(EPI_WIDTH-2*EPI_MARGIN));
    sprintf(vertical_zoom, "%f", SPGR_HEIGHT/(float)(EPI_HEIGHT-2*EPI_MARGIN));
    if(fork() == 0)
      {
      execl(izoom, izoom, epi_file, zoomed_epi_file, horizontal_zoom, vertical_zoom, "-i", (char *)0);
      fprintf(stderr, "Couldn't exec %s\n", izoom);
      exit(1);
      }
    wait((int *)0);
    read_spgr_image(zoomed_epi_file, zoomed_epi);
  /*calculate position of vertical meridian on SPGR image*/
    num_points = 0;
    for(y = 0; y != SPGR_HEIGHT; y++)
      shift_row(spgr[y], minpix, maxpix, 0.0, SPGR_WIDTH/2.0-SPGR_WIDTH/10.0, SPGR_WIDTH/2.0+SPGR_WIDTH/10.0, skew+num_points, stdev+num_points, &num_points);
    if(num_points < SPGR_HEIGHT/4)
      fprintf(stderr, "\nOnly %d of the %d scan lines yielded enough information to compute the skew.\nYou may want to specify the minimum and maximum intensities of brain pixels,\nas the final two parameters on the command line.\n", num_points, SPGR_HEIGHT);
    if(num_points == 0)
      exit(1);
    vertical_meridian = median(skew, num_points);
  /*calculate skewed vertical meridian on zoomed EPI image*/
    num_points = 0;
    for(y = 0; y != SPGR_HEIGHT; y++)
      {
      slice_num[num_points] = (float)y;
      shift_row(zoomed_epi[y], 32, 255, vertical_meridian, -SPGR_WIDTH/10.0, 0.0, skew+num_points, stdev+num_points, &num_points);
      }
  /*convert leftward (negative) skew values into positive values*/
    for(y = 0; y != num_points; y++)
      skew[y] *= -1.0;
  /*smooth the vector of skew values until it becomes nondecreasing*/
    while(smooth(skew, num_points))
      ;
  /*report all skew values*/
    printf("slice\tleftward skew\terror factor\n");
    for(x = 0; x != num_points; x++)
      printf("%.0f\t%f\t%f\n", slice_num[x], skew[x], stdev[x]);
  /*correct the averaged EPI image*/
    correct_image(zoomed_epi, slice_num, skew, num_points);
    write_spgr_image(unwarped_epi_file, zoomed_epi);
  /*correct the overlay image, which is warped in the same way as the EPI*/
    read_spgr_image(argv[5], spgr);
    correct_image(spgr, slice_num, skew, num_points);
    write_spgr_image(unwarped_overlay_file, spgr);
    }
  exit(0);
  }
