/*area.c - MRI measurement tool.
  Copyright (C) 1996, 2002 Matthew Belmonte
  Based on a problem definition by Stuart Hinds and Brian Egaas.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte.
  <mkb4@Cornell.edu>.  If you base a publication on data processed by this
  program, please notify Matthew Belmonte and acknowledge him in your
  publication.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/forms.h>
#include "sweet.h"

/*maximum number of different pixel values in a segmented image*/
#define MAX_NUM_TISSUE_CATEGORIES 16

static char izoom[] = "/usr/NeWS/bin/izoom",
	    standard_message[] = "press left button to sketch, middle button to draw lines, both to erase, right button to scroll";

FL_Coord display_width,		/*width in pixels of the server display*/
	 display_height;	/*height in pixels of the server display*/
double display_aspect_ratio;	/*aspect ratio of the server display (usu. 1)*/

FL_FORM *dialogue_form;		/*the master form*/
FL_OBJECT *graphics_obj,	/*the graphics window*/
	*message_obj,		/*the message window*/
	*horizontal_fov_obj,	/*horizontal field of view*/
	*vertical_fov_obj,	/*vertical field of view*/
	*zoom_obj,		/*magnification factor*/
	*label_obj,		/*data set label*/
	*label_export_obj,	/*toggle for export of data set label*/
	*area_obj,		/*area of measured region*/
	*(measurement_obj[MAX_NUM_TISSUE_CATEGORIES]), /*other measurements*/
	*histo_obj,		/*the histogram window*/
	*load_img_button,
	*next_img_button,
	*prev_img_button,
	*load_tracing_button,
	*next_tracing_button,
	*prev_tracing_button,
	*save_tracing_button,
	*clear_tracing_button,
	*measure_button,
	*export_select_button,
	*export_button,
	*quit_button;

Window graphics_window;		/*X window ID for graphics window*/
Colormap graphics_colourmap;	/*X colour map ID for graphics window*/
unsigned long tracing_colour;	/*foreground colour for tracings*/
unsigned long grey_level[0x100]; /*3-byte values for 256 colour codes*/
GC graphics_GC;			/*X graphics context for graphics window*/
int	graphics_window_width,	/*width in pixels of graphics window*/
	graphics_window_height,	/*height in pixels of graphics window*/
	xorigin,		/*coordinates of the window origin in the...*/
	yorigin;		/*...reference frame of the image*/
XImage *ximage;			/*X Image structure for graphics window*/
char	*copied_ximage_data,	/*image saved for restore_tracing()*/
	*tracing_data,		/*data area for tracing*/
	*copied_tracing_data;	/*tracing saved for restore_tracing()*/
unsigned char *eight_bit_data;	/*raw data area for graphics image*/
int	image_width,		/*width in pixels of unmagnified image*/
	image_height,		/*height in pixels of unmagnified image*/
	zoom;			/*magnification factor*/
float	horizontal_fov,		/*horizontal field of view (cm)*/
	vertical_fov;		/*vertical field of view (cm)*/
int	segmented,		/*1=segmented, 0=unsegmented*/
	tracing,		/*1=tracing is nonempty, 0=tracing is empty*/
	init_flag;		/*0 if image is loaded, 1 if initialising*/

int num_planes;			/*number of bitplanes*/
Window histo_window;		/*X window ID for histogram window*/
GC histo_GC;			/*X graphics context for histogram window*/
Visual *graphics_visual;	/*X Visual structure for graphics window*/
unsigned long histogram[256];	/*raw data for histogram*/
long histo_sum;			/*number of points in histogram*/
int histo_mean, histo_median, histo_mode, /*central tendency measures*/
	meas_mean, meas_median, meas_mode, /*copies for measurement data*/
	histo_display_height,	/*height in pixels of histogram window*/
	num_tissue_categories,	/*# of tissue categories in segmented image*/
	tissue_category_code[MAX_NUM_TISSUE_CATEGORIES];
double	area,			/*total area of region of interest*/
	measurement[MAX_NUM_TISSUE_CATEGORIES]; /*other measurements*/

int mouse_x, mouse_y;		/*mouse position in window coordinates*/
unsigned long mouse_evt_serial;	/*sequence number of most recent mouse event*/
enum {none, sketch, line, erase, scroll, measure} graphics_state; /*mouse act.*/
FILE *export_fp;		/*data file for exported measurements*/

/*get the dimensions and aspect ratio of the default display*/
int get_display_characteristics()
  {
  Display *display;
  int screen_num;
  display = XOpenDisplay((char *)0);
  if(display == (Display *)0)
    {
    fprintf(stderr, "setenv DISPLAY hostname:0.0 before running this program\n");
    return(1);
    }
  screen_num = DefaultScreen(display);
  display_width = DisplayWidth(display, screen_num);
  display_height = DisplayHeight(display, screen_num);
  display_aspect_ratio = DisplayHeightMM(display, screen_num)*display_width/(double)(DisplayWidthMM(display, screen_num)*display_height);
  XCloseDisplay(display);
  return(0);
  }

/*Display the image s.t. the window origin is at (xorigin, yorigin) in the
  reference frame of the image.  So if xorigin is negative, there's a border of
  unused space around the image, and if xorigin is positive, there's an
  unmapped fringe around the image - and similar for yorigin.*/
void display_image()
  {
  XPutImage(
    fl_display,
    graphics_window,
    graphics_GC,
    ximage,
    ((xorigin > 0)? xorigin: 0),
    ((yorigin > 0)? yorigin: 0),
    ((xorigin < 0)? -xorigin: 0),
    ((yorigin < 0)? -yorigin: 0),
    ((zoom*image_width-xorigin > graphics_window_width)? graphics_window_width: zoom*image_width-xorigin),
    ((zoom*image_height-yorigin > graphics_window_height)? graphics_window_height: zoom*image_height-yorigin));
  if(yorigin < 0)
    XClearArea(fl_display, graphics_window, 0, 0, graphics_window_width, -yorigin, False);
  if(zoom*image_height-yorigin < graphics_window_height)
    XClearArea(fl_display, graphics_window, 0, zoom*image_height-yorigin, graphics_window_width, graphics_window_height-zoom*image_height+yorigin, False);
  if(xorigin < 0)
    XClearArea(fl_display, graphics_window, 0, 0, -xorigin, graphics_window_height, False);
  if(zoom*image_width-xorigin < graphics_window_width)
    XClearArea(fl_display, graphics_window, zoom*image_width-xorigin, 0, graphics_window_width-zoom*image_width+xorigin, graphics_window_height, False);
  }

/*get the value of a point in the tracing - return 1 if set, 0 if clear*/
int get_tracing_point(img_x, img_y)
int img_x, img_y;
  {
  register int index;
  index = img_y*image_width*zoom+img_x;
  return((tracing_data[index/8]&(1<<(index%8))) != 0);
  }

/*set a point in the stored image, & set the corresponding point in the window*/
void set_tracing_point(img_x, img_y)
int img_x, img_y;
  {
  register index, win_x, win_y;
/*see if the point is already set*/
  index = img_y*image_width*zoom+img_x;
/*if not, set it*/
  if(!(tracing_data[index/8]&(1<<(index%8))))
    {
    tracing_data[index/8] |= 1<<(index%8);
  /*set the pixel in the image*/
    XPutPixel(ximage, img_x, img_y, tracing_colour);
  /*set the corresponding pixel on the display*/
    win_x = img_x-xorigin;
    win_y = img_y-yorigin;
    if((win_x >= 0) && (win_x < graphics_window_width) && (win_y >= 0) && (win_y < graphics_window_height))
      XDrawPoint(fl_display, graphics_window, graphics_GC, win_x, win_y);
  /*if this point begins a new tracing, reveal the tracing controls*/
    if(!tracing)
      {
      tracing = 1;
      fl_show_object(save_tracing_button);
      fl_show_object(clear_tracing_button);
      fl_show_object(measure_button);
      }
    }
  }

/*clear a point in the stored image and in the window*/
void clear_tracing_point(img_x, img_y)
int img_x, img_y;
  {
  register index, win_x, win_y;
/*see if the point is already clear*/
  index = img_y*image_width*zoom+img_x;
/*if not, clear it*/
  if(tracing_data[index/8]&(1<<(index%8)))
    {
    tracing_data[index/8] &= ~(1<<(index%8));
  /*clear that pixel in the image data...*/
    XPutPixel(ximage, img_x, img_y, grey_level[eight_bit_data[image_width*zoom*(image_height*zoom-1-img_y)+img_x]]);
  /*...and clear it in the displayed image*/
    win_x = img_x-xorigin;
    win_y = img_y-yorigin;
    if((win_x >= 0) && (win_x < graphics_window_width) && (win_y >= 0) && (win_y < graphics_window_height))
      XPutImage(fl_display, graphics_window, graphics_GC, ximage, img_x, img_y, win_x, win_y, 1, 1);
    }
  }

/*erase the stored tracing, and clear it from the stored image and the display*/
void clear_tracing()
  {
  register int x, y;
  for(y = 0; y != zoom*image_height; y++)
    for(x = 0; x != zoom*image_width; x++)
      clear_tracing_point(x, y);
  tracing = 0;
  }

/*re-establish consistency of the display and image data with the tracing data
  --similar to set_tracing_point() above*/
void display_tracing()
  {
  register int index, win_x, win_y, img_x, img_y;
  for(img_y = 0; img_y != zoom*image_height; img_y++)
    for(img_x = 0; img_x != zoom*image_width; img_x++)
      {
      index = img_y*image_width*zoom+img_x;
      if(tracing_data[index/8]&(1<<(index%8)))
	{
	XPutPixel(ximage, img_x, img_y, tracing_colour);
	win_x = img_x-xorigin;
	win_y = img_y-yorigin;
	if((win_x >= 0) && (win_x < graphics_window_width) && (win_y >= 0) && (win_y < graphics_window_height))
          XDrawPoint(fl_display, graphics_window, graphics_GC, win_x, win_y);
        }
      }
  }

/*draw a line between the given graphics window coordinates*/
void draw_line(x0, y0, x1, y1)
int x0, y0, x1, y1;
  {
  register double x, y, dx, dy, m;
  if((x1 != x0) || (y1 != y0))
    {
    if(x1 != x0)
      m = (y1-y0)/(double)(x1-x0);
    dx = (x0==x1)? 0.0: ((x1-x0)/abs(x1-x0))/sqrt(1.0+m*m);
    dy = (x0==x1)? (double)((y1-y0)/abs(y1-y0)): m*dx;
    for(x = x0, y = y0; ((x-x1)*(x-x0)<=0.0) && ((y-y1)*(y-y0)<=0.0); x += dx, y += dy)
      set_tracing_point(xorigin+(int)(0.5+x), yorigin+(int)(0.5+y));
    }
  set_tracing_point(x1+xorigin, y1+yorigin);
  }

/*erase along a line between the given graphics window coordinates*/
void erase_line(x0, y0, x1, y1, eraser_size)
int x0, y0, x1, y1, eraser_size;
  {
  register int x, y;
/*move along greatest dimension*/
  if(abs(y1-y0) > abs(x1-x0))
    {
  /*cap the vertical ends of the line*/
    y0 -= eraser_size*(y1-y0)/abs(y1-y0);
    if(y0 < 0)
      y0 = 0;
    else if(y0 >= zoom*image_height)
      y0 = zoom*image_height-1;
    y1 += (1+eraser_size)*(y1-y0)/abs(y1-y0);
    if(y1 < 0)
      y1 = 0;
    else if(y1 >= zoom*image_height)
      y1 = zoom*image_height-1;
  /*move vertically along the line, clearing the width of an eraser*/
    for(y = y0; y != y1; y += (y1-y0)/abs(y1-y0))
      for(x = x0+y*(x1-x0)/(y1-y0)-eraser_size/2; x <= x0+y*(x1-x0)/(y1-y0)+eraser_size/2; x++)
	if((x >= 0) && (x < zoom*image_width))
	  clear_tracing_point(x, y);
    }
  else
    {
  /*cap the horizontal ends of the line*/
    x0 -= eraser_size*(x1-x0)/abs(x1-x0);
    if(x0 < 0)
      x0 = 0;
    else if(x0 >= zoom*image_width)
      x0 = zoom*image_width-1;
    x1 += (1+eraser_size)*(x1-x0)/abs(x1-x0);
    if(x1 < 0)
      x1 = 0;
    else if(x1 >= zoom*image_width)
      x1 = zoom*image_width-1;
  /*move horizontally along the line, clearing the height of an eraser*/
    for(x = x0; x != x1; x += (x1-x0)/abs(x1-x0))
      for(y = y0+x*(y1-y0)/(x1-x0)-eraser_size/2; y <= y0+x*(y1-y0)/(x1-x0)+eraser_size/2; y++)
	if((y >= 0) && (y < zoom*image_height))
	  clear_tracing_point(x, y);
    }
  }

/*allocate memory for a tracing*/
void create_tracing()
  {
  register int i;
  tracing = 0;
  tracing_data = malloc((image_width*image_height*zoom*zoom+7)/8);
  if(tracing_data == (char *)0)
    {
    fprintf(stderr, "Out of memory\n");
    exit(1);
    }
  for(i = 0; i != (image_width*image_height*zoom*zoom+7)/8; i++)
    tracing_data[i] = 0;
  }

/*load the named tracing into memory, and display all on-screen portions*/
int load_tracing(name)
const char *name;
  {
  register int x, y;
  IMAGE *trace;
  unsigned short *rowbuf;
  if((trace = iopen(name, "r")) == (IMAGE *)0)
    {
    fprintf(stderr, "Couldn't open tracing file %s\n", name);
    tracing = 0;
    fl_hide_object(save_tracing_button);
    fl_hide_object(clear_tracing_button);
    fl_hide_object(measure_button);
    return(1);
    }
  if((trace->xsize != zoom*image_width) || (trace->ysize != zoom*image_height))
    {
    fprintf(stderr, "%dx%d tracing doesn't fit %dx%d zoomed image\n", trace->xsize, trace->ysize, zoom*image_width, zoom*image_height);
    iclose(trace);
    tracing = 0;
    fl_hide_object(save_tracing_button);
    fl_hide_object(clear_tracing_button);
    fl_hide_object(measure_button);
    return(1);
    }
  rowbuf = (unsigned short *)calloc(trace->xsize, sizeof(short));
  if(rowbuf == (unsigned short *)0)
    {
    fprintf(stderr, "Out of memory\n");
    tracing = 0;
    fl_hide_object(save_tracing_button);
    fl_hide_object(clear_tracing_button);
    fl_hide_object(measure_button);
    return(1);
    }
  for(y = 0; y != trace->ysize; y++)
    {
    getrow(trace, rowbuf, y, 0);
    for(x = 0; x != trace->xsize; x++)
      if(rowbuf[x])
	set_tracing_point(x, zoom*image_height-1-y);
    }
  free(rowbuf);
  iclose(trace);
  tracing = 1;
  fl_show_object(save_tracing_button);
  fl_show_object(clear_tracing_button);
  fl_show_object(measure_button);
  return(0);
  }

/*save a tracing as a black-and-white RLE file suitable for display and
  manipulation by other programs*/
void save_tracing(name)
const char *name;
  {
  register int x, y;
  unsigned short *rowbuf;
  IMAGE *img;
  rowbuf = (unsigned short *)calloc(zoom*image_width, sizeof(short));
  if(rowbuf == (unsigned short *)0)
    fprintf(stderr, "Couldn't save %s: out of memory\n", name);
  else
    {
    if((img = iopen(name, "w", RLE(1), 3, zoom*image_width, zoom*image_height, 1)) == (IMAGE *)0)
      fprintf(stderr, "Unable to write tracing file %s\n", name);
    else
      {
      for(y = 0; y != zoom*image_height; y++)
	{
	for(x = 0; x != zoom*image_width; x++)
	  rowbuf[x] = get_tracing_point(x, zoom*image_height-1-y)? 0xff: 0;
	putrow(img, rowbuf, y, 0);
	}
      iclose(img);
      }
    free(rowbuf);
    }
  }

/*save a copy of the current tracing for later use by restore_tracing()*/
void copy_tracing()
  {
  register int i;
  copied_tracing_data = tracing_data;
  create_tracing();
  for(i = 0; i != (image_width*image_height*zoom*zoom+7)/8; i++)
    tracing_data[i] = copied_tracing_data[i];
  copied_ximage_data = ximage->data;
  ximage->data = malloc(image_width*image_height*zoom*zoom*4);
  if(ximage->data == (char *)0)
    fprintf(stderr, "Out of memory\n");
  for(i = 0; i != image_width*image_height*zoom*zoom*4; i++)
    ximage->data[i] = copied_ximage_data[i];
  }

/*restore a tracing previously saved by copy_tracing()*/
void restore_tracing()
  {
  free(tracing_data);
  tracing_data = copied_tracing_data;
  free(ximage->data);
  ximage->data = copied_ximage_data;
  display_image();
  }

/*Redraw the histogram based on the pixel counts stored in 'histogram'.*/
void display_histogram()
  {
  register int intensity;

  XClearWindow(fl_display, histo_window);
  for(intensity = 0; intensity != 256; intensity++)
    XDrawLine(fl_display, histo_window, histo_GC, intensity, histo_display_height-1, intensity, histo_display_height-1-histogram[intensity]*(histo_display_height-1)/histogram[histo_mode]);
  }

/*Compute histo_mean, histo_median, and histo_mode from the current histogram.*/
void compute_histogram_central_tendency()
  {
  register int intensity;
  register long histogram_left_sum, intensity_sum;
  histo_sum = 0L;
  intensity_sum = 0L;
  histo_mode = 0;
  for(intensity = 0; intensity != 256; intensity++)
  /*inv: 'histo_sum' is the number of pixels having intensities less than
    'intensity'.  'intensity_sum' is the sum of the intensities of each of
    these pixels.  'histo_mode' is the most common intensity in the interval
    [0..intensity-1].*/
    {
    histo_sum += histogram[intensity];
    intensity_sum += histogram[intensity]*intensity;
    if(histogram[intensity] > histogram[histo_mode])
      histo_mode = intensity;
    }
  histo_mean = (int)((intensity_sum+histo_sum/2)/histo_sum);
  histo_median = -1;
  histogram_left_sum = 0L;
  while(histogram_left_sum <= histo_sum/2L)
  /*inv: 'histogram_left_sum' is the number of pixels having intensity less
    than or equal to 'histo_median'.*/
    histogram_left_sum += histogram[++histo_median];
  }

/*recursively fill the tracing-bounded area that includes the given image
  coordinates, and increment the appropriate elements of the histogram*/
void fill_area_and_set_histogram(img_x, img_y)
int img_x, img_y;
  {
  if((img_x < zoom*image_width) && (img_y < zoom*image_height) && (img_y >= 0))
    while((img_x >= 0) && !get_tracing_point(img_x, img_y))
      {
      histogram[eight_bit_data[(image_height*zoom-1-img_y)*image_width*zoom+img_x]]++;
      set_tracing_point(img_x, img_y);
      fill_area_and_set_histogram(img_x, 1+img_y);
      fill_area_and_set_histogram(img_x, img_y-1);
      fill_area_and_set_histogram(1+img_x--, img_y);
      }
  }

/*If the image is a segmented image, compute the total area of the measured
  region and the area of each tissue category within that region.  If the image
  is not segmented, compute the total area and report the mean, median, and
  modal intensities.*/
void compute_stats_from_histogram()
  {
  int tissue_category;
  double mm_per_pixel;
  char measurement_string[24];
/*compute area in square millimetres*/
  mm_per_pixel = atof(fl_get_input(horizontal_fov_obj))*atof(fl_get_input(vertical_fov_obj))*100.0/(zoom*zoom*image_width*image_height);
  area = histo_sum*mm_per_pixel;
  sprintf(measurement_string, "Area\n%.2f", area);
  fl_set_object_label(area_obj, measurement_string);
  fl_show_object(area_obj);
  if(segmented)
    for(tissue_category = 0; tissue_category != num_tissue_categories; tissue_category++)
      {
      measurement[tissue_category] = histogram[tissue_category_code[tissue_category]]*mm_per_pixel;
      sprintf(measurement_string, "(%d)\n%.2f", tissue_category_code[tissue_category], measurement[tissue_category]);
      fl_set_object_label(measurement_obj[tissue_category], measurement_string);
      fl_show_object(measurement_obj[tissue_category]);
      }
  else
    {
  /*length of text preceding numeric characters must be 5; see handler for
    export_obj*/
    sprintf(measurement_string, "mean\n%d", histo_mean);
    fl_set_object_label(measurement_obj[0], measurement_string);
    fl_show_object(measurement_obj[0]);
    sprintf(measurement_string, "med.\n%d", histo_median);
    fl_set_object_label(measurement_obj[1], measurement_string);
    fl_show_object(measurement_obj[1]);
    sprintf(measurement_string, "mode\n%d", histo_mode);
    fl_set_object_label(measurement_obj[2], measurement_string);
    fl_show_object(measurement_obj[2]);
    meas_mean = histo_mean;
    meas_median = histo_median;
    meas_mode = histo_mode;
    }
  }

/*measure the tracing-bounded area surrounding the given pixel*/
void measure_area(img_x, img_y)
  {
  register int intensity;
/*zero the histogram*/
  for(intensity = 0; intensity != 256; intensity++)
    histogram[intensity] = 0;
/*copy the current overlay so we can restore it after doing the fill*/
  copy_tracing();
/*recursively fill the area, and rewrite the histogram based on the pixels
  within the region of interest (unless the region of interest has zero area)*/
  fill_area_and_set_histogram(img_x, img_y);
  if(histo_sum != 0)
    {
    compute_histogram_central_tendency();
    display_histogram();
    compute_stats_from_histogram();
    sleep(1); /*maybe sleep 0 here if X runs on signals?*/
    }
/*restore the original overlay*/
  restore_tracing();
  fl_show_object(export_button);
  }

/*handle an exposure event in the graphics window*/
int graphics_expose(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  int	clr_x,	/*window coordinates of rectangle to clear*/
	clr_y,
	clr_h,	/*dimensions of rectangle to clear*/
	clr_w,
	put_x,	/*window coordinates of rectangle to re-image*/
	put_y,
	put_h,	/*dimensions of rectangle to re-image*/
	put_w;
  if(init_flag)
    XClearWindow(fl_display, graphics_window);
  else
    {
    if(xevt->xexpose.x >= -xorigin)
    /*left margin of redraw area is within image area*/
      {
      if(xevt->xexpose.x+xevt->xexpose.width <= zoom*image_width-xorigin)
      /*right margin of redraw area is within image area*/
	{
	put_x = xevt->xexpose.x;
	put_w = xevt->xexpose.width;
	if(xevt->xexpose.y >= -yorigin)
	/*top margin of redraw area is within image area*/
	  {
	  put_y = xevt->xexpose.y;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    {
	    clr_w = 0;
	    put_h = xevt->xexpose.height;
	    }
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    {
	    clr_x = xevt->xexpose.x;
	    clr_y = zoom*image_height-yorigin;
	    clr_w = xevt->xexpose.width;
	    clr_h = xevt->xexpose.y+xevt->xexpose.height-zoom*image_height+yorigin;
	    put_h = zoom*image_height-yorigin-xevt->xexpose.y;
	    }
	  }
	else
	/*top margin of redraw area is beyond image area*/
	  {
	  clr_x = xevt->xexpose.x;
	  clr_y = xevt->xexpose.y;
	  clr_w = xevt->xexpose.width;
	  put_y = -yorigin;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    {
	    clr_h = -yorigin-xevt->xexpose.y;
	    put_h = xevt->xexpose.height+xevt->xexpose.y+yorigin;
	    }
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    {
	    clr_h = xevt->xexpose.height;
	    put_h = zoom*image_height;
	    }
	  }
	}
      else
      /*right margin of redraw area is beyond image area*/
	{
	put_x = xevt->xexpose.x;
	put_w = zoom*image_width-xorigin-xevt->xexpose.x;
	if(xevt->xexpose.y >= -yorigin)
	/*top margin of redraw area is within image area*/
	  {
	  put_y = xevt->xexpose.y;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    {
	    clr_x = zoom*image_width-xorigin;
	    clr_y = xevt->xexpose.y;
	    clr_w = xevt->xexpose.width+xevt->xexpose.x+xorigin-zoom*image_width;
	    clr_h = xevt->xexpose.height;
	    put_h = xevt->xexpose.height;
	    }
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    {
	    clr_x = xevt->xexpose.x;
	    clr_y = xevt->xexpose.y;
	    clr_w = xevt->xexpose.width;
	    clr_h = xevt->xexpose.height;
	    put_h = zoom*image_height-yorigin-xevt->xexpose.y;
	    }
	  }
	else
	/*top margin of redraw area is beyond image area*/
	  {
	  clr_y = xevt->xexpose.y;
	  clr_w = xevt->xexpose.width;
	  clr_h = xevt->xexpose.height;
	  put_x = xevt->xexpose.x;
	  put_y = -yorigin;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    put_h = yorigin+xevt->xexpose.y+xevt->xexpose.height;
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    put_h = zoom*image_height;
	  }
	}
      }
    else
    /*left margin of redraw area is beyond image area*/
      {
      if(xevt->xexpose.x+xevt->xexpose.width <= zoom*image_width-xorigin)
      /*right margin of redraw area is within image area*/
	{
	clr_x = xevt->xexpose.x;
	clr_y = xevt->xexpose.y;
	clr_h = xevt->xexpose.height;
	put_x = -xorigin;
	put_w = xorigin+xevt->xexpose.x+xevt->xexpose.width;
	if(xevt->xexpose.y >= -yorigin)
	/*top margin of redraw area is within image area*/
	  {
	  put_y = xevt->xexpose.y;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    {
	    clr_w = -xorigin-xevt->xexpose.x;
	    put_h = xevt->xexpose.height;
	    }
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    {
	    clr_w = xevt->xexpose.width;
	    put_h = zoom*image_height-yorigin-xevt->xexpose.y;
	    }
	  }
	else
	/*top margin of redraw area is beyond image area*/
	  {
	  clr_w = xevt->xexpose.width;
	  put_y = -yorigin;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    put_h = yorigin+xevt->xexpose.y+xevt->xexpose.height;
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    put_h = zoom*image_height;
	  }
	}
      else
      /*right margin of redraw area is beyond image area*/
	{
	clr_x = xevt->xexpose.x;
	clr_y = xevt->xexpose.y;
	clr_w = xevt->xexpose.width;
	clr_h = xevt->xexpose.height;
	if(xevt->xexpose.y >= -yorigin)
	/*top margin of redraw area is within image area*/
	  {
	  put_x = -xorigin;
	  put_y = xevt->xexpose.y;
	  put_w = zoom*image_width;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    put_h = xevt->xexpose.height;
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    put_h = zoom*image_height-yorigin-xevt->xexpose.y;
	  }
	else
	/*top margin of redraw area is beyond image area*/
	  {
	  put_x = -xorigin;
	  put_y = -yorigin;
	  put_w = zoom*image_width;
	  if(xevt->xexpose.y+height <= zoom*image_height-yorigin)
	  /*bottom margin of redraw area is within image area*/
	    put_h = yorigin+xevt->xexpose.y+xevt->xexpose.height;
	  else
	  /*bottom margin of redraw area is beyond image area*/
	    put_h = zoom*image_height;
	  }
	}
      }
    if(clr_w != 0)
      XClearArea(fl_display, graphics_window, clr_x, clr_y, clr_w, clr_h, False);
    XPutImage(fl_display, graphics_window, graphics_GC, ximage, put_x+xorigin, put_y+yorigin, put_x, put_y, put_w, put_h);
    }
  return(0);
  }

/*scroll the image by adjusting the relative window origin and redrawing*/
void scroll_image(dx, dy)
int dx, dy;
  {
  xorigin -= dx;
  yorigin -= dy;
  display_image(); /*tried XCopyArea() and graphics_expose(), but didn't work*/
  }

/*handle an exposure generated by scrolling as if it were an ordinary exposure*/
int graphics_gexpose(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  XEvent expose_evt;
  expose_evt.xexpose.x = xevt->xgraphicsexpose.x;
  expose_evt.xexpose.y = xevt->xgraphicsexpose.y;
  expose_evt.xexpose.width = xevt->xgraphicsexpose.width;
  expose_evt.xexpose.height = xevt->xgraphicsexpose.height;
  return(graphics_expose(object, win, width, height, &expose_evt, dummy));
  }

/*handle a button press within the graphics window*/
int button_down_handler(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  if(xevt->xbutton.serial <= mouse_evt_serial)
    {
    fprintf(stderr, "missed button event\n");
    graphics_state = none;
    }
  else
    {
    mouse_evt_serial = xevt->xbutton.serial;
    mouse_x = (xevt->xbutton.x<0)? 0: (xevt->xbutton.x>=graphics_window_width)? graphics_window_width-1: xevt->xbutton.x;
    mouse_y = (xevt->xbutton.y<0)? 0: (xevt->xbutton.y>=graphics_window_height)? graphics_window_height-1: xevt->xbutton.y;
    switch(graphics_state)
      {
      case none:
	switch(xevt->xbutton.button)
	  {
	  case 1:
	    graphics_state = sketch;
	    break;
	  case 2:
	    graphics_state = line;
	    break;
	  case 3:
	    graphics_state = scroll;
	  }
	break;
      case sketch:
	if(xevt->xbutton.button == 2)
	  graphics_state = erase;
	break;
      case line:
	if(xevt->xbutton.button == 1)
	  graphics_state = erase;
	break;
      case measure:
	if(xevt->xbutton.button == 1)
	  {
	  fl_set_object_label(message_obj, "measuring..."); XFlush(fl_display);
	  measure_area(mouse_x+xorigin, mouse_y+yorigin);
	  }
	graphics_state = none;
	fl_set_object_label(message_obj, standard_message);
      }
    }
  return(0);
  }

/*handle the release of a button within the graphics window*/
int button_up_handler(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  if(xevt->xbutton.serial <= mouse_evt_serial)
    {
    fprintf(stderr, "missed button event\n");
    graphics_state = none;
    }
  else
    {
    mouse_evt_serial = xevt->xbutton.serial;
    switch(graphics_state)
      {
      case line:
	if(xevt->xbutton.button == 2)
	  {
	  if((mouse_x >= 0) && (mouse_x < graphics_window_width)
	   && (mouse_y >= 0) && (mouse_y < graphics_window_height)
	   && (xevt->xbutton.x >= 0) && (xevt->xbutton.x < graphics_window_width)
	   && (xevt->xbutton.y >= 0) && (xevt->xbutton.y < graphics_window_height))
	    draw_line(mouse_x, mouse_y, xevt->xbutton.x, xevt->xbutton.y);
	  graphics_state = none;
	  }
	break;
      case sketch:
	if(xevt->xbutton.button == 1)
	  graphics_state = none;
	break;
      case erase:
	if((xevt->xbutton.button == 1) || (xevt->xbutton.button == 2))
	  graphics_state = none;
	break;
      case scroll:
	if(xevt->xbutton.button == 3)
	  graphics_state = none;
      }
    }
  return(0);
  }

/*handle mouse motion while a button is down within the graphics window*/
int mouse_handler(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  switch(graphics_state)
    {
    case sketch:
      if((mouse_x >= 0) && (mouse_x < graphics_window_width)
       && (mouse_y >= 0) && (mouse_y < graphics_window_height)
       && (xevt->xbutton.x >= 0) && (xevt->xbutton.x < graphics_window_width)
       && (xevt->xbutton.y >= 0) && (xevt->xbutton.y < graphics_window_height)
       && (mouse_x >= -xorigin) && (mouse_x < zoom*image_width-xorigin)
       && (mouse_y >= -yorigin) && (mouse_y < zoom*image_height-yorigin)
       && (xevt->xbutton.x >= -xorigin) && (xevt->xbutton.x < zoom*image_width-xorigin)
       && (xevt->xbutton.y >= -yorigin) && (xevt->xbutton.y < zoom*image_height-yorigin))
	draw_line(mouse_x, mouse_y, xevt->xmotion.x, xevt->xmotion.y);
      mouse_x = xevt->xmotion.x;
      mouse_y = xevt->xmotion.y;
      break;
    case erase:
      if((xevt->xmotion.x != mouse_x) || (xevt->xmotion.y != mouse_y))
	{
	erase_line(mouse_x+xorigin, mouse_y+yorigin, xevt->xmotion.x+xorigin, xevt->xmotion.y+yorigin, zoom|1);
	mouse_x = xevt->xmotion.x;
	mouse_y = xevt->xmotion.y;
	}
      break;
    case scroll:
      {
      int dx, dy;
      dx = xevt->xmotion.x-mouse_x;
      dy = xevt->xmotion.y-mouse_y;
      if(xorigin+graphics_window_width-dx > zoom*image_width)
	dx = xorigin+graphics_window_width-zoom*image_width;
      if(yorigin+graphics_window_height-dy > zoom*image_height)
	dy = yorigin+graphics_window_height-zoom*image_height;
      if(xorigin < 0)
	dx = 0;
      else if(xorigin-dx < 0)
	dx = xorigin;
      if(yorigin < 0)
	dy = 0;
      else if(yorigin-dy < 0)
	dy = yorigin;
      if(dx || dy)
	scroll_image(dx, dy);
      mouse_x = xevt->xmotion.x;
      mouse_y = xevt->xmotion.y;
      }
    }
  return(0);
  }

int histo_expose(object, win, width, height, xevt, dummy)
FL_OBJECT *object;
Window win;
int width, height;
XEvent *xevt;
void *dummy;
  {
  register int intensity;
  if(init_flag)
    XClearWindow(fl_display, histo_window);
  else
    {
    XClearArea(fl_display, histo_window, xevt->xexpose.x, xevt->xexpose.y, xevt->xexpose.width, xevt->xexpose.height, False);
    for(intensity = xevt->xexpose.x; intensity != xevt->xexpose.x+xevt->xexpose.width; intensity++)
      XDrawLine(fl_display, histo_window, histo_GC, intensity, histo_display_height-1, intensity, histo_display_height-1-histogram[intensity]*(histo_display_height-1)/histogram[histo_mode]);
    }
  return(0);
  }

/*initialise all forms and associated data structures*/
void create_forms()
  {
  FL_Coord button_height;
  int i;
  button_height = display_height/17.0;
  dialogue_form = fl_bgn_form(FL_NO_BOX, display_width, display_height);
  fl_add_box(FL_FLAT_BOX, 0.0, 0.0, display_width, display_height, "");
  message_obj = fl_add_text(FL_NORMAL_TEXT, 0.0, display_height-button_height, display_width, button_height, "initialising...");
  fl_set_object_lsize(message_obj, 18.0);
  horizontal_fov_obj = fl_add_input(FL_FLOAT_INPUT, display_width-171.0, display_height-button_height-button_height/2.0, 43.0, button_height/2.0, "Width (cm)");
  fl_set_object_lsize(horizontal_fov_obj, 8.0);
  vertical_fov_obj = fl_add_input(FL_FLOAT_INPUT, display_width-43.0, display_height-button_height-button_height/2.0, 43.0, button_height/2.0, "Height (cm)");
  fl_set_object_lsize(vertical_fov_obj, 8.0);
  zoom_obj = fl_add_input(FL_INT_INPUT, display_width-171.0, display_height-2.0*button_height, 171.0, button_height/2.0, "Magnification");
  fl_set_object_lsize(zoom_obj, 8.0);
  label_obj = fl_add_input(FL_NORMAL_INPUT, display_width-192.0, display_height-3.0*button_height, 128.0, button_height, "Label:");
  fl_set_object_lsize(label_obj, 12.0);
  fl_hide_object(label_obj);
  label_export_obj = fl_add_checkbutton(FL_PUSH_BUTTON, display_width-256.0, display_height-3.0*button_height, 32.0, button_height, "");
  fl_hide_object(label_export_obj);
  area_obj = fl_add_lightbutton(FL_PUSH_BUTTON, display_width-64.0, display_height-3.0*button_height, 64.0, button_height, "Area");
  fl_set_object_lsize(area_obj, 12.0);
  fl_hide_object(area_obj);
  for(i = 0; i != MAX_NUM_TISSUE_CATEGORIES; i++)
    {
    measurement_obj[i] = fl_add_lightbutton(FL_PUSH_BUTTON, display_width-64.0*(4-i%4), display_height-(4+i/4)*button_height, 64.0, button_height, "");
    fl_set_object_lsize(measurement_obj[i], 12.0);
    fl_hide_object(measurement_obj[i]);
    }
  histo_obj = fl_add_canvas(FL_NORMAL_CANVAS, display_width-256.0, display_height-13.0*button_height, 256.0, 6.0*button_height, "");
  histo_display_height = (int)(6.0*button_height);
  next_img_button = fl_add_button(FL_NORMAL_BUTTON, display_width-256.0, display_height-14.0*button_height, 85.0, button_height, "Next Image");
  fl_set_object_lsize(next_img_button, 12.0);
  fl_hide_object(next_img_button);
  prev_img_button = fl_add_button(FL_NORMAL_BUTTON, display_width-171.0, display_height-14.0*button_height, 86.0, button_height, "Previous Image");
  fl_set_object_lsize(prev_img_button, 12.0);
  fl_hide_object(prev_img_button);
  load_img_button = fl_add_button(FL_NORMAL_BUTTON, display_width-85.0, display_height-14.0*button_height, 85.0, button_height, "Load Image");
  fl_set_object_lsize(load_img_button, 12.0);
  next_tracing_button = fl_add_button(FL_NORMAL_BUTTON, display_width-256.0, display_height-15.0*button_height, 85.0, button_height, "Next Tracing");
  fl_set_object_lsize(next_tracing_button, 12.0);
  fl_hide_object(next_tracing_button);
  prev_tracing_button = fl_add_button(FL_NORMAL_BUTTON, display_width-171.0, display_height-15.0*button_height, 86.0, button_height, "Previous Tracing");
  fl_set_object_lsize(prev_tracing_button, 12.0);
  fl_hide_object(prev_tracing_button);
  load_tracing_button = fl_add_button(FL_NORMAL_BUTTON, display_width-85.0, display_height-15.0*button_height, 85.0, button_height, "Load Tracing");
  fl_set_object_lsize(load_tracing_button, 12.0);
  fl_hide_object(load_tracing_button);
  save_tracing_button = fl_add_button(FL_NORMAL_BUTTON, display_width-256.0, display_height-16.0*button_height, 85.0, button_height, "Save Tracing");
  fl_set_object_lsize(save_tracing_button, 12.0);
  fl_hide_object(save_tracing_button);
  clear_tracing_button = fl_add_button(FL_NORMAL_BUTTON, display_width-171.0, display_height-16.0*button_height, 86.0, button_height, "Clear Tracing");
  fl_set_object_lsize(clear_tracing_button, 12.0);
  fl_hide_object(clear_tracing_button);
  quit_button = fl_add_button(FL_NORMAL_BUTTON, display_width-85.0, display_height-16.0*button_height, 85.0, button_height, "Quit");
  fl_set_object_lsize(quit_button, 18.0);
  measure_button = fl_add_button(FL_NORMAL_BUTTON, display_width-256.0, display_height-17.0*button_height, 85.0, button_height, "Measure");
  fl_set_object_lsize(measure_button, 12.0);
  fl_hide_object(measure_button);
  export_button = fl_add_button(FL_NORMAL_BUTTON, display_width-171.0, display_height-17.0*button_height, 86.0, button_height, "Save\nMeasurement");
  fl_set_object_lsize(export_button, 12.0);
  fl_hide_object(export_button);
  export_select_button = fl_add_button(FL_NORMAL_BUTTON, display_width-85.0, display_height-17.0*button_height, 85.0, button_height, "Select\nMeasurement\nData File");
  fl_set_object_lsize(export_select_button, 12.0);
  graphics_obj = fl_add_canvas(FL_NORMAL_CANVAS, 0.0, 0.0, display_width-256.0, display_height-button_height, "");
  graphics_window_width = display_width-256;
  graphics_window_height = display_height-button_height;
  fl_end_form();
  }

int init_graphics_windows()
  {
  register unsigned short grey, sixteen_bit_grey;
  XColor colour_def;
  XVisualInfo visual_info;
/*customise the graphics window*/
  num_planes = 24;
  if(!XMatchVisualInfo(fl_display, fl_screen, num_planes, TrueColor, &visual_info))
    while((num_planes != 0) && !XMatchVisualInfo(fl_display, fl_screen, num_planes, StaticColor, &visual_info))
      num_planes -= 8;
  if(num_planes == 0)
    {
    fprintf(stderr, "This program requires a colour display.\n");
    return(1);
    }
  printf("using %d bitplanes\n", num_planes); fflush(stdout);
  graphics_visual = visual_info.visual;
  graphics_colourmap = fl_create_colormap(&visual_info, 0);
  fl_set_canvas_visual(graphics_obj, graphics_visual);
  fl_set_canvas_depth(graphics_obj, num_planes);
  fl_set_canvas_colormap(graphics_obj, graphics_colourmap);
/*customise the histogram window*/
  fl_set_canvas_visual(histo_obj, graphics_visual);
  fl_set_canvas_depth(histo_obj, num_planes);
  fl_set_canvas_colormap(histo_obj, graphics_colourmap);
/*allocate 8-bit grey levels*/
  colour_def.flags = 0;
  sixteen_bit_grey = 0;
  for(grey = 0; grey != 0x100; grey++)
    {
    colour_def.red = colour_def.green = colour_def.blue = sixteen_bit_grey;
    if(!XAllocColor(fl_display, graphics_colourmap, &colour_def))
      fprintf(stderr, "Couldn't allocate grey level %d\n", grey);
    grey_level[grey] = colour_def.pixel;
    sixteen_bit_grey += 0x100;
    }
/*add a hot pink foreground colour for overlays*/
  colour_def.red = 0xffff;
  colour_def.green = 0x7fff;
  colour_def.blue = 0x7fff;
  if(!XAllocColor(fl_display, graphics_colourmap, &colour_def))
    fprintf(stderr, "Couldn't allocate overlay colour\n");
  tracing_colour = colour_def.pixel;
  return(0);
  }

/*Read the RLE image of the given name.  Record its width, its height, and a
  pointer to its data stored in row-major order.  Return 0 if successful, 1 if
  error.*/
int read_rle_image(name, width, height)
const char *name;
int *width, *height;
  {
  register int col, row;
  unsigned short *rowbuf;
  IMAGE *img;
  img = iopen(name, "r");
  if(img == (IMAGE *)0)
    {
    fprintf(stderr, "Couldn't open RLE image %s\n", name);
    return(1);
    }
  *width = img->xsize;
  *height = img->ysize;
  eight_bit_data = (unsigned char *)malloc(*width**height);
  if(eight_bit_data == (unsigned char *)0)
    {
    iclose(img);
    fprintf(stderr, "Out of memory\n");
    return(1);
    }
  rowbuf = (unsigned short *)calloc(*width, sizeof(short));
  if(rowbuf == (unsigned short *)0)
    {
    iclose(img);
    fprintf(stderr, "Out of memory\n");
    return(1);
    }
  for(row = 0; row != *height; row++)
    {
    getrow(img, rowbuf, row, 0);
    for(col = 0; col != *width; col++)
      eight_bit_data[row**width+col] = (unsigned char)(rowbuf[col]&0xff);
    }
  free(rowbuf);
  iclose(img);
  return(0);
  }

/*Use the given eight-bit uncompressed binary image data to create an X image
  whose width is image_width*zoom and whose height is image_height*zoom.  Make
  ximage point to the XImage structure that describes this new image.  Return 0
  if successful, 1 if error.*/
int create_ximage()
  {
  register int i, j;
  char *ximage_data;
/*maximum of 4 bytes per pixel*/
  ximage_data = malloc(image_width*image_height*zoom*zoom*4);
  if(ximage_data == (char *)0)
    {
    fprintf(stderr, "Out of memory\n");
    ximage = (XImage *)0;
    return(1);
    }
  ximage = XCreateImage(fl_display, graphics_visual, num_planes, ZPixmap, 0, ximage_data, image_width*zoom, image_height*zoom, 8, image_width*zoom*4);
  if(ximage == (XImage *)0)
    {
    fprintf(stderr, "Couldn't create X image\n");
    return(1);
    }
  for(i = 0; i != image_height*zoom; i++)
    for(j = 0; j != image_width*zoom; j++)
      {
    /*set the pixel*/
      XPutPixel(ximage, j, image_height*zoom-1-i, grey_level[eight_bit_data[image_width*zoom*i+j]]);
      }
  return(0);
  }

/*Load the image stored in the file of the given name, and compute its
  intensity histogram.  If there are sixteen or fewer distinct intensity levels,
  assume that this is a segmented image, and set the 'segmented' flag.
  Otherwise, clear the 'segmented' flag.  Set the 'width' annd 'height'
  parameters appropriately.*/
int load_image_and_compute_histogram(name)
const char *name;
  {
  register int col, row;
  if(read_rle_image(name, &image_width, &image_height))
    return(1);
  for(col = 0; col != 256; col++)
    histogram[col] = 0;
  for(row = 0; row != image_height; row++)
    for(col = 0; col != image_width; col++)
      histogram[eight_bit_data[row*image_width+col]]++;
  row = 0;
  for(col = 0; col != 256; col++)
    row += (histogram[col] != 0);
  segmented = (row <= MAX_NUM_TISSUE_CATEGORIES);
/*if the image is segmented, set tissue categories and label export buttons*/
  if(segmented)
    {
    num_tissue_categories = 0;
    for(col = 0; col != 256; col++)
      if(histogram[col] != 0)
	tissue_category_code[num_tissue_categories++] = col;
    }
  compute_histogram_central_tendency();
  return(0);
  }

/*Zoom the imaged stored in the file of the given name according to the
  magnification factor stored in 'zoom', and then load the result.  If the
  'segmented' flag is clear, blur the zoomed image with a triangular filter.
  Verify that the dimensions of the zoomed image are the product of the
  magnification factor and the original dimensions given by 'image_width' and
  'image_height'.*/
int load_zoomed_image(name)
const char *name;
  {
  int zoomed_width, zoomed_height;
  char temp_file_name[24], magnification[8], filter_type[3];
  sprintf(temp_file_name, "/tmp/area%d.bw", (int)(getpid()));
  sprintf(magnification, "%d", zoom);
  filter_type[0] = '-';
  filter_type[1] = (segmented? 'i': 't');
  filter_type[2] = '\0';
  if(fork() == 0)
    {
    execl(izoom, izoom, name, temp_file_name, magnification, magnification, filter_type, (char *)0);
    fprintf(stderr, "Couldn't execute %s\n", izoom);
    return(1);
    }
  wait((int *)0);
  if(read_rle_image(temp_file_name, &zoomed_width, &zoomed_height))
    return(1);
  unlink(temp_file_name);
  if((zoom*image_width != zoomed_width) || (zoom*image_height != zoomed_height))
    {
    fprintf(stderr, "Zoom inconsistency (%dx%d image zoomed %d times yields %dx%d)\n", image_width, image_height, zoom, zoomed_width, zoomed_height);
    return(1);
    }
  return(0);
  }

/*Load the named image, display it at the current magnification, and compute
  and display its intensity histogram.*/
int load_image(name)
const char *name;
  {
  fl_set_object_label(message_obj, "loading..."); XFlush(fl_display);
  if(load_image_and_compute_histogram(name) || ((zoom > 1) && load_zoomed_image(name)))
    return(1);
  xorigin = (zoom*image_width-graphics_window_width)/2;
  yorigin = (zoom*image_height-graphics_window_height)/2;
  if(create_ximage())
    return(1);
  display_image();
  display_histogram();
  return(0);
  }

/*Destroy the currently stored image*/
void destroy_image()
  {
  register char *ximage_data;
  ximage_data = ximage->data;
  XDestroyImage(ximage);
  free(ximage_data);
  free(eight_bit_data);
  }

/*split a pathname into a directory path and a filename*/
void get_path_components(dir, file, path)
char *dir, *file;
const char *path;
  {
  register int i, separator_position;
  separator_position = -1;
  for(i = 0; path[i] != '\0'; i++)
    if(path[i] == '/')
      separator_position = i;
  for(i = 0; i < separator_position; i++)
    dir[i] = path[i];
  dir[i++] = '\0';
  file -= i;
  do
    file[i] = path[i];
  while(path[i++] != '\0');
  }

/*query user for a name for the export data file, and open it for writing*/
void select_export_file(meas_dir, meas_file)
char *meas_dir, *meas_file;
  {
  const char *meas_path;
  meas_path = fl_show_fselector("select data file to store exported measurements", meas_dir, "*.csv", "");
  if(meas_path != (char *)0)
    {
    export_fp = fopen(meas_path, "a");
    if(export_fp == NULL)
      fprintf(stderr, "Couldn't open export file %s\n", meas_path);
    else
      {
      fl_show_object(label_export_obj);
      get_path_components(meas_dir, meas_file, meas_path);
      }
    }
  }

/*If the given filename contains a numeric substring, and the file whose name
  is the given filename substituted with a numeric substring that represents a
  decimal integer one greater than that represented by the original substring
  exists and is readable, set 'next_file' to the name of this file, and show
  the 'next_button' object.  If not, set 'next_file' to the null string.  Act
  similarly for the previous filename, with a decrement of the original
  number.*/
void show_buttons_if_adjacent_images_available(dir, file, next_file, prev_file, next_button, prev_button)
const char *dir, *file;
char *next_file, *prev_file;
FL_OBJECT *next_button, *prev_button;
  {
  register int i, file_num, filename_prefix_len;
  register char *pathname_ptr;
  IMAGE *img;
  char pathname[MAXPATHLEN];
/*copy directory name into pathname*/
  pathname_ptr = pathname;
  i = 0;
  while(dir[i] != '\0')
    *(pathname_ptr++) = dir[i++];
/*add separator*/
  *(pathname_ptr++) = '/';
/*copy filename prefix*/
  i = 0;
  while((file[i] != '\0') && ((file[i] < '0') || (file[i] > '9')))
    {
    if(next_file != (char *)0)
      next_file[i] = file[i];
    if(prev_file != (char *)0)
      prev_file[i] = file[i];
    i++;
    }
  if(file[i] == '\0')
    {
    if(next_file != (char *)0)
      *next_file = '\0';
    if(prev_file != (char *)0)
      *prev_file = '\0';
    }
  else
  /*scan a number from the filename*/
    {
    filename_prefix_len = i;
    file_num = 0;
    while((file[i] >= '0') && (file[i] <= '9'))
      file_num = 10*file_num+file[i++]-'0';
  /*copy filename suffix and attempt to open image files*/
    if(next_file != (char *)0)
      {
      sprintf(next_file+filename_prefix_len, "%d", file_num+1);
      strcat(next_file+filename_prefix_len, file+i);
      strcpy(pathname_ptr, next_file);
      if((img = iopen(pathname, "r")) != (IMAGE *)0)
	{
	fl_show_object(next_button);
	iclose(img);
	}
      else
	{
	fl_hide_object(next_button);
	*next_file = '\0';
	}
      }
    if(prev_file != (char *)0)
      {
      sprintf(prev_file+filename_prefix_len, "%d", file_num-1);
      strcat(prev_file+filename_prefix_len, file+i);
      strcpy(pathname_ptr, prev_file);
      if((img = iopen(pathname, "r")) != (IMAGE *)0)
	{
	fl_show_object(prev_button);
	iclose(img);
	}
      else
	{
	fl_hide_object(prev_button);
	*prev_file = '\0';
	}
      }
    }
  }

int main(argc, argv)
int argc;
char **argv;
  {
  int i, prev_zoom, prev_width, prev_height;
  const char *home_dir;
  FILE *prefs;
  FL_OBJECT *obj;
  XGCValues gcvals;
  XSetWindowAttributes attrs;
  const char *image_path;
  char	image_dir[MAXPATHLEN],
	image_file[MAXPATHLEN],
	next_img_file[MAXPATHLEN],
	prev_img_file[MAXPATHLEN],
	trace_dir[MAXPATHLEN],
	trace_file[MAXPATHLEN],
	next_trace_file[MAXPATHLEN],
	prev_trace_file[MAXPATHLEN],
	meas_dir[MAXPATHLEN],
	meas_file[MAXPATHLEN],
	prefs_file[MAXPATHLEN],
	buf[MAXPATHLEN];
  home_dir = getenv("HOME");
  if((home_dir != (const char *)0)
   &&((prefs = fopen(strcat(strcpy(prefs_file, home_dir), "/.area"), "r")) != NULL))
    {
    /*read settings from preferences file*/
    if((i = fscanf(prefs, "%f %f %d %s %s %s %s", &horizontal_fov, &vertical_fov, &zoom, image_dir, trace_dir, meas_dir, meas_file)) < 6)
      {
      fprintf(stderr, "invalid format in preferences file %s\n", prefs_file);
      horizontal_fov = 20.0;
      vertical_fov = 20.0;
      zoom = 4;
      getwd(image_dir);
      strcpy(trace_dir, image_dir);
      strcpy(meas_dir, image_dir);
      meas_file[0] = '\0';
      }
    else if(i == 6)
      meas_file[0] = '\0';
    fclose(prefs);
    }
  else
  /*set default preferences*/
    {
    horizontal_fov = 20.0;
    vertical_fov = 20.0;
    zoom = 4;
    if(!getwd(image_dir))
      {
      fprintf(stderr, "Can't stat .\n");
      exit(1);
      }
    strcpy(trace_dir, image_dir);
    strcpy(meas_dir, image_dir);
    meas_file[0] = '\0';
    }
  trace_file[0] = '*';
  trace_file[1] = '\0';
  if(get_display_characteristics())
    exit(1);
  if((display_aspect_ratio >= 1.05) || (display_aspect_ratio <= 0.95))
    {
    fprintf(stderr, "Displays with non-square pixels are not supported.\n");
    exit(1);
    }
/*allow space for borders*/
  display_width -= 16.0;
  display_height -= 32.0;
/*set up graphical user interface*/
  fl_initialize(&argc, argv, "area", 0, 0); /*OLD VERSION WAS fl_initialize(argv[0], "area", 0, 0, &argc, argv);*/
  create_forms();
  sprintf(buf, "%g", horizontal_fov);
  fl_set_input(horizontal_fov_obj, buf);
  sprintf(buf, "%g", vertical_fov);
  fl_set_input(vertical_fov_obj, buf);
  sprintf(buf, "%d", zoom);
  fl_set_input(zoom_obj, buf);
  init_flag = 1;
  tracing = 0;
  if(init_graphics_windows())
    exit(1);
  fl_show_form(dialogue_form, FL_PLACE_SIZE, FL_NOBORDER, "MRI Measurement Tool");
  graphics_window = fl_get_canvas_id(graphics_obj);
  histo_window = fl_get_canvas_id(histo_obj);
/*set graphics background colour*/
  attrs.background_pixel = grey_level[0x50];
  XChangeWindowAttributes(fl_display, graphics_window, CWBackPixel, &attrs);
/*set histogram background colour*/
  attrs.background_pixel = grey_level[0];
  XChangeWindowAttributes(fl_display, histo_window, CWBackPixel, &attrs);
/*set tracing colour*/
  gcvals.foreground = tracing_colour;
  gcvals.graphics_exposures = True;
  graphics_GC = XCreateGC(fl_display, graphics_window, GCForeground|GCGraphicsExposures, &gcvals);
/*set histogram foreground colour*/
  gcvals.foreground = grey_level[0xff];
  histo_GC = XCreateGC(fl_display, histo_window, GCForeground, &gcvals);
/*set up redrawing routines*/
  fl_add_canvas_handler(graphics_obj, Expose, graphics_expose, (void *)0);
  fl_add_canvas_handler(graphics_obj, GraphicsExpose, graphics_gexpose, (void *)0);
  fl_add_canvas_handler(histo_obj, Expose, histo_expose, (void *)0);
/*set up other event handlers*/
  graphics_state = none;
  mouse_evt_serial = 0L;
  fl_add_selected_xevent(graphics_window, ButtonPressMask|ButtonReleaseMask|Button1MotionMask|Button3MotionMask|OwnerGrabButtonMask);
  fl_add_canvas_handler(graphics_obj, ButtonPress, button_down_handler, (void *)0);
  fl_add_canvas_handler(graphics_obj, ButtonRelease, button_up_handler, (void *)0);
  fl_add_canvas_handler(graphics_obj, MotionNotify, mouse_handler, (void *)0);
/*initially, there is no data file for exported measurements*/
  export_fp = NULL;
/*process events*/
  fl_set_object_label(message_obj, "area (C) 1996 by Matthew Belmonte - proper acknowledgement must be made in publications");
  while((obj = fl_do_forms()) != quit_button)
    if(obj == load_img_button)
      {
      image_path = fl_show_fselector("select an image file", image_dir, "*.bw", "");
      if(image_path != (char *)0)
	{
	if(!init_flag)
	  {
	  init_flag = 1;
	  prev_zoom = zoom;
	  prev_width = image_width;
	  prev_height = image_height;
	  destroy_image();
	  }
	else
	  {
	  prev_zoom = -1;
	  prev_width = 1;
	  prev_height = 1;
	  }
	zoom = atoi(fl_get_input(zoom_obj));
	if(zoom < 1)
	  zoom = 1;
	if(load_image(image_path))
	  {
	  fl_hide_object(load_tracing_button);
	  fl_hide_object(next_img_button);
	  fl_hide_object(prev_img_button);
	  }
	else
	  {
	  init_flag = 0;
	  if((prev_width*prev_zoom != image_width*zoom) || (prev_height*prev_zoom != image_height*zoom))
	    {
	    free(tracing_data);
	    create_tracing();
	    fl_hide_object(clear_tracing_button);
	    fl_hide_object(save_tracing_button);
	    }
	  else
	    display_tracing();
	  fl_show_object(load_tracing_button);
	  get_path_components(image_dir, image_file, image_path);
	/*write the name of the current image file on the Load Image button*/
	  sprintf(buf, "Image\n%s", image_file);
	  fl_set_object_label(load_img_button, buf);
	/*copy the name of the current image file into the data set label*/
	  fl_set_input(label_obj, image_file);
	  fl_show_object(label_obj);
	  show_buttons_if_adjacent_images_available(image_dir, image_file, next_img_file, prev_img_file, next_img_button, prev_img_button);
	  }
	}
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == next_img_button)
      {
      prev_width = image_width;
      prev_height = image_height;
      prev_zoom = zoom;
      zoom = atoi(fl_get_input(zoom_obj));
      if(zoom < 1)
	zoom = 1;
      for(i = 0; image_dir[i] != '\0'; i++)
	buf[i] = image_dir[i];
      buf[i++] = '/';
      destroy_image();
      sprintf(buf+i, "%s", next_img_file);
      load_image(buf);
      if((prev_width*prev_zoom != image_width*zoom) || (prev_height*prev_zoom != image_height*zoom))
	{
	free(tracing_data);
	create_tracing();
	}
      else
	display_tracing();
      fl_show_object(prev_img_button);
      strcpy(prev_img_file, image_file);
      strcpy(image_file, next_img_file);
      sprintf(buf, "Image\n%s", image_file);
      fl_set_object_label(load_img_button, buf);
      show_buttons_if_adjacent_images_available(image_dir, image_file, next_img_file, (char *)0, next_img_button, (FL_OBJECT *)0);
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == prev_img_button)
      {
      prev_width = image_width;
      prev_height = image_height;
      prev_zoom = zoom;
      zoom = atoi(fl_get_input(zoom_obj));
      if(zoom < 1)
	zoom = 1;
      for(i = 0; image_dir[i] != '\0'; i++)
	buf[i] = image_dir[i];
      buf[i++] = '/';
      destroy_image();
      sprintf(buf+i, "%s", prev_img_file);
      load_image(buf);
      if((prev_width*prev_zoom != image_width*zoom) || (prev_height*prev_zoom != image_height*zoom))
	{
	free(tracing_data);
	create_tracing();
	}
      else
	{
	display_tracing();
	}
      fl_show_object(next_img_button);
      strcpy(next_img_file, image_file);
      strcpy(image_file, prev_img_file);
      sprintf(buf, "Image\n%s", image_file);
      fl_set_object_label(load_img_button, buf);
      show_buttons_if_adjacent_images_available(image_dir, image_file, (char *)0, prev_img_file, (FL_OBJECT *)0, prev_img_button);
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == load_tracing_button)
      {
      image_path = fl_show_fselector("select a tracing file:", trace_dir, "*.tr", "");
      if(image_path != (char *)0)
	{
	clear_tracing();
	if(!load_tracing(image_path))
	  {
	  get_path_components(trace_dir, trace_file, image_path);
	  show_buttons_if_adjacent_images_available(trace_dir, trace_file, next_trace_file, prev_trace_file, next_tracing_button, prev_tracing_button);
	  }
	}
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == next_tracing_button)
      {
      clear_tracing();
      for(i = 0; trace_dir[i] != '\0'; i++)
	buf[i] = trace_dir[i];
      buf[i++] = '/';
      sprintf(buf+i, "%s", next_trace_file);
      load_tracing(buf);
      fl_show_object(prev_tracing_button);
      get_path_components(trace_dir, trace_file, image_path);
      strcpy(prev_trace_file, trace_file);
      strcpy(trace_file, next_trace_file);
      show_buttons_if_adjacent_images_available(trace_dir, trace_file, next_trace_file, (char *)0, next_tracing_button, (FL_OBJECT *)0);
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == prev_tracing_button)
      {
      clear_tracing();
      for(i = 0; trace_dir[i] != '\0'; i++)
	buf[i] = trace_dir[i];
      buf[i++] = '/';
      sprintf(buf+i, prev_trace_file);
      load_tracing(buf);
      fl_show_object(next_tracing_button);
      get_path_components(trace_dir, trace_file, image_path);
      strcpy(next_trace_file, trace_file);
      strcpy(trace_file, prev_trace_file);
      show_buttons_if_adjacent_images_available(trace_dir, trace_file, (char *)0, prev_trace_file, (FL_OBJECT *)0, prev_tracing_button);
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == save_tracing_button)
      {
      image_path = fl_show_fselector("select a name for the saved tracing:", trace_dir, "*", "");
      if(image_path != (char *)0)
	{
	save_tracing(image_path);
	get_path_components(trace_dir, trace_file, image_path);
	fl_hide_object(next_tracing_button);
	fl_hide_object(prev_tracing_button);
	show_buttons_if_adjacent_images_available(trace_dir, trace_file, next_trace_file, prev_trace_file, next_tracing_button, prev_tracing_button);
	fl_set_object_label(message_obj, standard_message);
	}
      }
    else if(obj == clear_tracing_button)
      {
      clear_tracing();
      fl_hide_object(save_tracing_button);
      fl_hide_object(clear_tracing_button);
      fl_hide_object(measure_button);
      fl_set_object_label(message_obj, standard_message);
      }
    else if(obj == measure_button)
      {
      fl_set_object_label(message_obj, "click left button inside the region to be measured, or click any other button to cancel");
      graphics_state = measure;
      }
    else if(obj == export_select_button)
      select_export_file(meas_dir, meas_file);
    else if(obj == export_button)
      {
      if(export_fp == NULL)
	select_export_file(meas_dir, meas_file);
      if(export_fp != NULL)
	{
	if(fl_get_button(label_export_obj))
	  fprintf(export_fp, "%s,", fl_get_input(label_obj));
	if(fl_get_button(area_obj))
	  fprintf(export_fp, "%.2f,", area);
	if(segmented)
	  {
	  for(i = 0; i != num_tissue_categories; i++)
	    if(fl_get_button(measurement_obj[i]))
	      fprintf(export_fp, "%.2f,", measurement[i]);
	  }
	else
	  {
	  if(fl_get_button(measurement_obj[0]))
	    fprintf(export_fp, "%d,", meas_mean);
	  if(fl_get_button(measurement_obj[1]))
	    fprintf(export_fp, "%d,", meas_median);
	  if(fl_get_button(measurement_obj[2]))
	    fprintf(export_fp, "%d,", meas_mode);
	  }
	putc('\n', export_fp);
	fflush(export_fp); /*save measurement in case something bad happens*/
	}
      }
/*close the measurements file*/
  if(export_fp != NULL)
    fclose(export_fp);
/*save settings in the preferences file*/
  if((home_dir != (const char *)0) && ((prefs = fopen(prefs_file, "w")) != NULL))
    {
    fprintf(prefs, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n", fl_get_input(horizontal_fov_obj), fl_get_input(vertical_fov_obj), fl_get_input(zoom_obj), image_dir, trace_dir, meas_dir, meas_file);
    fclose(prefs);
    }
  return 0;
  }
