/*epicat.c - concatenate Advanced NMR echo-planar image files
  Copyright (c) 2000 Matthew Belmonte

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  If you find this program useful, please send mail to Matthew Belmonte
  <mkb4@Cornell.edu>.
*/

#define FOOTER_LEN 16	/*length of the footer on each image*/

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>

/*Read parameters from the .irp file "<prefix>.irp".  Return 1 if the
  parameters are read successfully and match master copies, 0 otherwise.*/
int read_parameters(prefix, images_per_slice, num_slices, xsize, ysize, master_fp)
const char *prefix;	/*prefix of the .irp file name*/
int *images_per_slice;	/*images per slice value is written through this ptr*/
int *num_slices,	/*these data are written if master_fp, or*/
    *xsize, *ysize;	/*compared if !master_fp*/
FILE **master_fp;	/*non-null iff this .irp file is being used as a
			  template and the file should be left open, null
			  otherwise*/
  {
  int x, y, n;
  FILE *irp;
  char filename[MAXPATHLEN];
  for(n = 0; prefix[n] != '\0'; n++)
    filename[n] = prefix[n];
  strcpy(filename+n, ".irp");
  irp = fopen(filename, "r");
  if(irp == NULL)
    {
    perror(filename);
    return 0;
    }
  if(fscanf(irp, "num_slices %d images_per_slice %d xsize %d ysize %d", &n, images_per_slice, &x, &y) != 4)
    {
    fclose(irp);
    fprintf(stderr, "%s: bad format - file omitted\n", filename);
    return 0;
    }
  if(master_fp)
    {
    *master_fp = irp;
    *num_slices = n;
    *xsize = x;
    *ysize = y;
    }
  else
    {
    fclose(irp);
    if((n != *num_slices) || (x != *xsize) || (y != *ysize))
      {
      fprintf(stderr, "%s dimensions (%d,%d,%d) don't match master dimensions (%d,%d,%d) -\nfile omitted", filename, x, y, n, *xsize, *ysize, *num_slices);
      return 0;
      }
    }
  return 1;
  }

/*Use shell globbing to find an existing file whose name is of the form
  <prefix><s>.<ext>, where <s> is the lexicographically least string possible.
  Return 1 if a name of this form exists, 0 otherwise.*/
int glob(prefix, ext, result)
const char *prefix, *ext;		/*filename and extension*/
char *result;
  {
  FILE *shell;
  int files;
  char cmd[MAXPATHLEN+4];
  sprintf(cmd, "ls %s*.%s", prefix, ext);
  shell = popen(cmd, "r");
  files = fscanf(shell, "%s", result);
  pclose(shell);
  return(files == 1);
    /*the shell will have taken care of printing the error message for us*/
  }

/*Write a file "<prefix>.irp" that specifies the given values for
  images_per_slice, num_slices, xsize, and ysize, and whose remaining
  parameters are read from the current position in the file master_irp.*/
void write_output_irp(prefix, images_per_slice, num_slices, xsize, ysize, master_irp)
const char *prefix;
int images_per_slice, num_slices, xsize, ysize;
FILE *master_irp;
  {
  register int i;
  FILE *irp;
  char filename[MAXPATHLEN];
  for(i = 0; prefix[i] != '\0'; i++)
    filename[i] = prefix[i];
  strcpy(filename+i, ".irp");
  irp = fopen(filename, "w");
  if(irp == NULL)
    perror(filename);
  else
    {
    fprintf(irp, "num_slices    %d\nimages_per_slice    %d\nxsize    %d\nysize    %d\n", num_slices, images_per_slice, xsize, ysize);
    if((i = getc(master_irp)) != EOF)
      {
      if(i != '\n')
	putc(i, irp);
      while((i = getc(master_irp)) != EOF)
	putc(i, irp);
      }
    fclose(irp);
    }
  }

/*Return 1 if size of the lexicographically least element of the set
  "<prefix>*.img" names a file whose size matches the given spatial and
  temporal dimensions, 0 otherwise.*/
int check_img_size(prefix, images_per_slice, num_slices, xsize, ysize)
const char *prefix;
int images_per_slice, num_slices, xsize, ysize;
  {
  struct stat stbuf;
  char filename[MAXPATHLEN];
  if(!glob(prefix, "img", filename))
    return 0;
  if(stat(filename, &stbuf))
    {
    perror(filename);
    return 0;
    }
  if(stbuf.st_size != num_slices*images_per_slice*(2L*ysize*xsize + FOOTER_LEN))
    {
    fprintf(stderr, "%s size %ld not equal to %d slices *\n%d images per slice * (%d bytes header + %d*%d * 2 bytes per pixel\n- image omitted\n", filename, stbuf.st_size, num_slices, images_per_slice, FOOTER_LEN, xsize, ysize);
    return 0;
    }
  return 1;
  }

/*Open an output file with a ".img" extension.  Return 1 if successful, 0
  otherwise.*/
int open_output_img(prefix, fpp)
const char *prefix;
FILE **fpp;
  {
  register int i;
  char filename[MAXPATHLEN];
  for(i = 0; prefix[i] != '\0'; i++)
    filename[i] = prefix[i];
  strcpy(filename+i, ".img");
  *fpp = fopen(filename, "wb");
  if(*fpp == NULL)
    {
    perror(filename);
    return 0;
    }
  return 1;
  }

int main(argc, argv)
int argc;
char **argv;
  {
  register int img, slice, study, i;
  int *images_per_slice, total_images_per_slice, num_slices, xsize, ysize;
  FILE *master_irp, *output_irp, *input_img, *output_img;
  char *imgbuf, filename[MAXPATHLEN];
  if(argc < 3)
    {
    fprintf(stderr, "usage: %s <source path> ... <source path> <dest. path>\nFor each <source path>, <source path>.img or <source path>_001.img and\n<source_path>.irp should exist.  Destination files with the corresponding\nsuffixes will be created, and any existing files with the same names as the\ndestination files will be overwritten.\n", *argv);
    exit(1);
    }
  images_per_slice = (int *)calloc(argc-2, sizeof(*images_per_slice));
  if(images_per_slice == (int *)0)
    {
    perror((char *)0);
    exit(errno);
    }
/*get dimensions of first input file, to use as a template*/
  if(!read_parameters(argv[1], images_per_slice, &num_slices, &xsize, &ysize, &master_irp))
    exit(1);
/*allocate buffer for image data*/
  imgbuf = (char *)calloc(xsize*ysize, 2);
  if(imgbuf == (char *)0)
    {
    perror((char *)0);
    exit(errno);
    }
/*open output image file*/
  if(!open_output_img(argv[argc-1], &output_img))
    {
    fclose(master_irp);
    exit(1);
    }
  images_per_slice--; /*index from 1 rather than 0*/
/*check file sizes against the dimensions specified in the .irp files*/
  for(study = 1; study != argc-1; study++)
    if(!read_parameters(argv[study], images_per_slice+study, &num_slices, &xsize, &ysize, (FILE *)0)
    || !check_img_size(argv[study], images_per_slice[study], num_slices, xsize, ysize))
      images_per_slice[study] = 0;
/*for each slice, step through all the input files*/
  for(slice = 1; slice <= num_slices; slice++)
    {
    total_images_per_slice = 0;
    for(study = 1; study != argc-1; study++)
      {
      if(images_per_slice[study])
	{
      /*glob the filename and open it*/
	glob(argv[study], "img", filename);
	input_img = fopen(filename, "rb");
	if(input_img == NULL)
	  perror(filename);
	else
	  {
	/*seek to the current slice within this study*/
	  fseek(input_img, (slice-1)*images_per_slice[study]*(2L*xsize*ysize + FOOTER_LEN), SEEK_SET);
	/*copy all images from the current slice in the current study*/
	  for(img = 1; img <= images_per_slice[study]; img++)
	    {
	    total_images_per_slice++;
	  /*copy the image*/
	    fread(imgbuf, 1, xsize*ysize*2, input_img);
	    fwrite(imgbuf, 1, xsize*ysize*2, output_img);
	  /*copy the slice number*/
	    putc(getc(input_img), output_img);
	    putc(getc(input_img), output_img);
	  /*write the new image number*/
	    putc((total_images_per_slice>>8)&0xff, output_img);
	    putc(total_images_per_slice & 0xff, output_img);
	    fseek(input_img, 2L, SEEK_CUR);
	  /*copy the rest of the footer*/
	    for(i = 4; i != FOOTER_LEN; i++)
	      putc(getc(input_img), output_img);
	    }
	  fclose(input_img);
	  }
	}
      }
    }
  free(imgbuf);
  free(images_per_slice+1);
  fclose(output_img);
/*write the new parameters file*/
  write_output_irp(argv[argc-1], total_images_per_slice, num_slices, xsize, ysize, master_irp);
  fclose(master_irp);
  return 0;
  }
